#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2023


"""
operations related to Processings.
"""

import datetime

from idds.orm.base.session import read_session, transactional_session
from idds.common.constants import ProcessingLocking, ProcessingStatus, GranularityType, ContentRelationType
from idds.orm import (processings as orm_processings,
                      collections as orm_collections,
                      contents as orm_contents,
                      messages as orm_messages,
                      transforms as orm_transforms)


@transactional_session
def add_processing(request_id, workload_id, transform_id, status, submitter=None,
                   substatus=ProcessingStatus.New, granularity=None,
                   granularity_type=GranularityType.File,
                   new_poll_period=1, update_poll_period=10,
                   new_retries=0, update_retries=0, max_new_retries=3, max_update_retries=0,
                   expired_at=None, processing_metadata=None, session=None):
    """
    Add a processing.

    :param request_id: The request id.
    :param workload_id: The workload id.
    :param transform_id: Transform id.
    :param status: processing status.
    :param submitter: submitter name.
    :param granularity: Granularity size.
    :param granularity_type: Granularity type.
    :param expired_at: The datetime when it expires.
    :param processing_metadata: The metadata as json.

    :raises DuplicatedObject: If a processing with the same name exists.
    :raises DatabaseException: If there is a database error.

    :returns: processing id.
    """
    return orm_processings.add_processing(request_id=request_id, workload_id=workload_id, transform_id=transform_id,
                                          status=status, substatus=substatus, submitter=submitter,
                                          granularity=granularity, granularity_type=granularity_type,
                                          new_poll_period=new_poll_period,
                                          update_poll_period=update_poll_period,
                                          new_retries=new_retries, update_retries=update_retries,
                                          max_new_retries=max_new_retries,
                                          max_update_retries=max_update_retries,
                                          expired_at=expired_at, processing_metadata=processing_metadata,
                                          session=session)


@read_session
def get_processing(processing_id=None, to_json=False, session=None):
    """
    Get processing or raise a NoObject exception.

    :param processing_id: Processing id.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no processing is founded.

    :returns: Processing.
    """
    return orm_processings.get_processing(processing_id=processing_id, to_json=to_json, session=session)


@read_session
def get_processings(request_id=None, workload_id=None, transform_id=None, to_json=False, session=None):
    """
    Get processing or raise a NoObject exception.

    :param processing_id: Processing id.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no processing is founded.

    :returns: Processing.
    """
    return orm_processings.get_processings(request_id=request_id, workload_id=workload_id,
                                           transform_id=transform_id, to_json=to_json, session=session)


@read_session
def get_processings_by_transform_id(transform_id=None, to_json=False, session=None):
    """
    Get processings or raise a NoObject exception.

    :param tranform_id: Transform id.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no processing is founded.

    :returns: Processings.
    """
    return orm_processings.get_processings_by_transform_id(transform_id=transform_id, to_json=to_json, session=session)


@transactional_session
def get_processing_by_id_status(processing_id, status=None, locking=False, session=None):
    pr = orm_processings.get_processing_by_id_status(processing_id=processing_id, status=status, locking=locking, session=session)
    if pr is not None and locking:
        parameters = {}
        parameters['locking'] = ProcessingLocking.Locking
        parameters['updated_at'] = datetime.datetime.utcnow()
        orm_processings.update_processing(processing_id=pr['processing_id'], parameters=parameters, session=session)
    return pr


@transactional_session
def get_processings_by_status(status, time_period=None, locking=False, bulk_size=None, to_json=False, by_substatus=False,
                              not_lock=False, next_poll_at=None, for_poller=False, only_return_id=False,
                              locking_for_update=False, new_poll=False, update_poll=False, session=None):
    """
    Get processing or raise a NoObject exception.

    :param status: Processing status of list of processing status.
    :param time_period: Time period in seconds.
    :param locking: Whether to retrieve only unlocked items and lock them.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no processing is founded.

    :returns: Processings.
    """
    if locking:
        if not only_return_id and bulk_size:
            # order by cannot work together with locking. So first select 2 * bulk_size without locking with order by.
            # then select with locking.
            proc_ids = orm_processings.get_processings_by_status(status=status, period=time_period, locking=locking,
                                                                 bulk_size=bulk_size * 2, to_json=False, locking_for_update=False,
                                                                 by_substatus=by_substatus, only_return_id=True,
                                                                 for_poller=for_poller, new_poll=new_poll,
                                                                 update_poll=update_poll, session=session)
            if proc_ids:
                processing2s = orm_processings.get_processings_by_status(status=status, period=time_period, locking=locking,
                                                                         processing_ids=proc_ids,
                                                                         bulk_size=None, to_json=to_json,
                                                                         locking_for_update=locking_for_update,
                                                                         by_substatus=by_substatus, only_return_id=only_return_id,
                                                                         new_poll=new_poll, update_poll=update_poll,
                                                                         for_poller=for_poller, session=session)
                if processing2s:
                    # reqs = req2s[:bulk_size]
                    # order requests
                    processings = []
                    for proc_id in proc_ids:
                        if len(processings) >= bulk_size:
                            break
                        for p in processing2s:
                            if p['processing_id'] == proc_id:
                                processings.append(p)
                                break
                    # processings = processings[:bulk_size]
                else:
                    processings = []
            else:
                processings = []
        else:
            processings = orm_processings.get_processings_by_status(status=status, period=time_period, locking=locking,
                                                                    bulk_size=bulk_size, to_json=to_json,
                                                                    locking_for_update=locking_for_update,
                                                                    new_poll=new_poll, update_poll=update_poll,
                                                                    only_return_id=only_return_id,
                                                                    by_substatus=by_substatus, for_poller=for_poller, session=session)

        parameters = {}
        if not not_lock:
            parameters['locking'] = ProcessingLocking.Locking
        if next_poll_at:
            parameters['next_poll_at'] = next_poll_at
        parameters['updated_at'] = datetime.datetime.utcnow()
        if parameters:
            for processing in processings:
                if type(processing) in [dict]:
                    orm_processings.update_processing(processing['processing_id'], parameters=parameters, session=session)
                else:
                    orm_processings.update_processing(processing, parameters=parameters, session=session)
    else:
        processings = orm_processings.get_processings_by_status(status=status, period=time_period, locking=locking,
                                                                bulk_size=bulk_size, to_json=to_json,
                                                                new_poll=new_poll, update_poll=update_poll,
                                                                only_return_id=only_return_id,
                                                                by_substatus=by_substatus, for_poller=for_poller, session=session)
    return processings


@transactional_session
def update_processing(processing_id, parameters, session=None):
    """
    update a processing.

    :param processing_id: the transform id.
    :param parameters: A dictionary of parameters.
    :param session: The database session in use.

    :raises NoObject: If no content is founded.
    :raises DatabaseException: If there is a database error.

    """
    return orm_processings.update_processing(processing_id=processing_id, parameters=parameters, session=session)


@transactional_session
def delete_processing(processing_id=None, session=None):
    """
    delete a processing.

    :param processing_id: The id of the processing.
    :param session: The database session in use.

    :raises NoObject: If no processing is founded.
    :raises DatabaseException: If there is a database error.
    """
    return orm_processings.delete_processing(processing_id=processing_id, session=session)


@transactional_session
def update_processing_with_collection_contents(updated_processing, new_processing=None, updated_collection=None,
                                               updated_files=None, new_files=None,
                                               coll_msg_content=None, file_msg_content=None, transform_updates=None,
                                               message_bulk_size=1000, session=None):
    """
    Update processing with collection, contents, file messages and collection messages.

    :param updated_processing: dict with processing id and parameters.
    :param updated_collection: dict with collection id and parameters.
    :param updated_files: list of content files.
    :param coll_msg_content: message with collection info.
    :param file_msg_content: message with files info.
    """
    if updated_files:
        orm_contents.update_contents(updated_files, session=session)
    if new_files:
        orm_contents.add_contents(contents=new_files, session=session)
    if file_msg_content:
        if not type(file_msg_content) in [list, tuple]:
            file_msg_content = [file_msg_content]
        for file_msg_con in file_msg_content:
            orm_messages.add_message(msg_type=file_msg_con['msg_type'],
                                     status=file_msg_con['status'],
                                     source=file_msg_con['source'],
                                     transform_id=file_msg_con['transform_id'],
                                     num_contents=file_msg_con['num_contents'],
                                     msg_content=file_msg_con['msg_content'],
                                     bulk_size=message_bulk_size,
                                     session=session)
    if updated_collection:
        orm_collections.update_collection(coll_id=updated_collection['coll_id'],
                                          parameters=updated_collection['parameters'],
                                          session=session)
    if coll_msg_content:
        orm_messages.add_message(msg_type=coll_msg_content['msg_type'],
                                 status=coll_msg_content['status'],
                                 source=coll_msg_content['source'],
                                 transform_id=coll_msg_content['transform_id'],
                                 num_contents=coll_msg_content['num_contents'],
                                 msg_content=coll_msg_content['msg_content'],
                                 session=session)
    if updated_processing:
        orm_processings.update_processing(processing_id=updated_processing['processing_id'],
                                          parameters=updated_processing['parameters'],
                                          session=session)
    if new_processing:
        orm_processings.add_processing(**new_processing, session=session)
    if transform_updates:
        orm_transforms.update_transform(transform_id=transform_updates['transform_id'],
                                        parameters=transform_updates['parameters'],
                                        session=session)


def resolve_input_dependency_id(new_input_dependency_contents, session=None):
    coll_ids = []
    for content in new_input_dependency_contents:
        if content['coll_id'] not in coll_ids:
            coll_ids.append(content['coll_id'])
    contents = orm_contents.get_contents(coll_id=coll_ids, relation_type=ContentRelationType.Output, session=session)
    content_name_id_map = {}
    for content in contents:
        if content['coll_id'] not in content_name_id_map:
            content_name_id_map[content['coll_id']] = {}
        content_name_id_map[content['coll_id']][content['name']] = content['content_id']
    for content in new_input_dependency_contents:
        content_dep_id = content_name_id_map[content['coll_id']][content['name']]
        content['content_dep_id'] = content_dep_id
    return new_input_dependency_contents


@transactional_session
def update_processing_contents(update_processing, update_contents, update_messages=None, new_contents=None,
                               update_dep_contents=None, update_collections=None, messages=None,
                               new_update_contents=None, new_input_dependency_contents=None,
                               new_contents_ext=None, update_contents_ext=None,
                               message_bulk_size=2000, session=None):
    """
    Update processing with contents.

    :param update_processing: dict with processing id and parameters.
    :param update_contents: list of content files.
    """
    if update_collections:
        orm_collections.update_collections(update_collections, session=session)
    if update_contents:
        orm_contents.update_contents(update_contents, session=session)
    if new_update_contents:
        # first add and then delete, to trigger the trigger 'update_content_dep_status'.
        # too slow
        orm_contents.add_contents_update(new_update_contents, session=session)
        # orm_contents.delete_contents_update(session=session)
        pass
    if new_contents:
        orm_contents.add_contents(new_contents, session=session)
    if new_contents_ext:
        orm_contents.add_contents_ext(new_contents_ext, session=session)
    if update_contents_ext:
        orm_contents.update_contents_ext(update_contents_ext, session=session)
    if new_input_dependency_contents:
        new_input_dependency_contents = resolve_input_dependency_id(new_input_dependency_contents, session=session)
        orm_contents.add_contents(new_input_dependency_contents, session=session)
    if update_dep_contents:
        request_id, update_dep_contents_status_name, update_dep_contents_status = update_dep_contents
        for status_name in update_dep_contents_status_name:
            status = update_dep_contents_status_name[status_name]
            status_content_ids = update_dep_contents_status[status_name]
            if status_content_ids:
                orm_contents.update_dep_contents(request_id, status_content_ids, status, session=session)
    if update_processing:
        orm_processings.update_processing(processing_id=update_processing['processing_id'],
                                          parameters=update_processing['parameters'],
                                          session=session)
    if update_messages:
        orm_messages.update_messages(update_messages, bulk_size=message_bulk_size, session=session)
    if messages:
        if not type(messages) in [list, tuple]:
            messages = [messages]
        orm_messages.add_messages(messages, bulk_size=message_bulk_size, session=session)


@transactional_session
def clean_locking(time_period=3600, session=None):
    """
    Clearn locking which is older than time period.

    :param time_period in seconds
    """
    orm_processings.clean_locking(time_period=time_period, session=session)


@transactional_session
def clean_next_poll_at(status, session=None):
    """
    Clearn next_poll_at.

    :param status: status of the processing
    """
    orm_processings.clean_next_poll_at(status=status, session=session)



#!/usr/bin/env python

from idds.common.constants import MessageStatus                        # noqa F401
from idds.core.messages import retrieve_messages, add_messages         # noqa F401


def clean_msg(msg):
    new_msg = {'msg_type': msg['msg_type'],
               'status': MessageStatus.New,
               'source': msg['source'],
               'destination': msg['destination'],
               'request_id': msg['request_id'],
               'workload_id': msg['workload_id'],
               'transform_id': msg['transform_id'],
               'num_contents': msg['num_contents'],
               'msg_content': msg['msg_content']}
    return new_msg


for req_id in range(1073, 1084):
    msgs = retrieve_messages(request_id=req_id)
    number_contents = 0
    for msg in msgs:
        # print(json_dumps(msg['msg_content'], sort_keys=True, indent=4))
        # if msg['num_contents'] > 10000 and msg['status'] == MessageStatus.New:
        if msg['num_contents'] >= 10000:
            print(msg['request_id'], msg['msg_id'], msg['num_contents'])
            msg = clean_msg(msg)
            add_messages([msg], bulk_size=10000)
        # break



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2022 - 2023


"""
Test client.
"""

import os
import tarfile
import time
import uuid

# from pandaclient import Client

# import traceback

# from rucio.client.client import Client as Rucio_Client
# from rucio.common.exception import CannotAuthenticate

# from idds.client.client import Client
from idds.client.clientmanager import ClientManager
# from idds.common.constants import RequestType, RequestStatus
from idds.common.utils import get_rest_host
# from idds.tests.common import get_example_real_tape_stagein_request
# from idds.tests.common import get_example_prodsys2_tape_stagein_request

# from idds.workflowv2.work import Work, Parameter, WorkStatus
# from idds.workflowv2.workflow import Condition, Workflow
from idds.workflowv2.workflow import Workflow
# from idds.atlas.workflowv2.atlasstageinwork import ATLASStageinWork
from idds.doma.workflowv2.domapandawork import DomaPanDAWork

# task_cloud = 'LSST'
task_cloud = 'US'

task_queue = 'DOMA_LSST_GOOGLE_TEST'
# task_queue = 'DOMA_LSST_GOOGLE_MERGE'
# task_queue = 'SLAC_TEST'
# task_queue = 'DOMA_LSST_SLAC_TEST'
task_queue = 'SLAC_Rubin'
# task_queue = 'CC-IN2P3_TEST'


os.environ['PANDA_URL'] = 'http://pandaserver-doma.cern.ch:25080/server/panda'
os.environ['PANDA_URL_SSL'] = 'https://pandaserver-doma.cern.ch:25443/server/panda'
os.environ["PANDACACHE_URL"] = os.environ["PANDA_URL_SSL"]


def get_local_pfns():
    working_dir = os.path.dirname(os.path.realpath(__file__))
    local_pfns = []
    files = ['test_domapanda.py',
             'test_domapanda_build.py']
    for f in files:
        full_filename = os.path.join(working_dir, f)
        if os.path.exists(full_filename):
            print("Adding: %s" % full_filename)
            local_pfns.append(full_filename)
        else:
            print("Not exist: %s" % full_filename)
    # add iDDS client code, to be compiled on worker node.
    base_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(working_dir))))
    for idds_dir in ['client', 'common', 'workflow', 'doma']:
        idds_full_dir = os.path.join(base_dir, idds_dir)
        local_pfns.append(idds_full_dir)
    # add main test script
    cmd_build = os.path.join(base_dir, 'main/tools/rubin/cmdline_builder.py')
    local_pfns.append(cmd_build)
    main_script = os.path.join(base_dir, 'main/tools/rubin/test_build.sh')
    local_pfns.append(main_script)
    return local_pfns


def create_archive_file(output_filename, local_pfns):
    if not output_filename.startswith("/"):
        output_filename = os.path.join('/tmp/wguan', output_filename)

    with tarfile.open(output_filename, "w:gz", dereference=True) as tar:
        for local_pfn in local_pfns:
            base_name = os.path.basename(local_pfn)
            tar.add(local_pfn, arcname=os.path.basename(base_name))
    return output_filename


def copy_files_to_pandacache(filename):
    from pandaclient import Client
    status, out = Client.putFile(filename, True)
    print("copy_files_to_pandacache: status: %s, out: %s" % (status, out))
    if out.startswith("NewFileName:"):
        # found the same input sandbox to reuse
        filename = out.split(":")[-1]
    elif out != "True":
        print(out)
        return None

    filename = os.path.basename(filename)
    cache_path = os.path.join(os.environ["PANDACACHE_URL"], 'cache')
    filename = os.path.join(cache_path, filename)
    return filename


def setup_workflow():
    local_pfns = get_local_pfns()
    output_filename = "jobO.%s.tar.gz" % str(uuid.uuid4())
    output_filename = create_archive_file(output_filename, local_pfns)
    print("archive file: %s" % output_filename)
    output_filename = copy_files_to_pandacache(output_filename)
    print("pandacache file: %s" % output_filename)

    # executable = 'unset PYTHONPATH; source /cvmfs/sw.lsst.eu/linux-x86_64/lsst_distrib/w_2022_53/loadLSST.bash; pwd; ls -al; setup lsst_distrib;\n python3 ${CTRL_BPS_PANDA_DIR}/python/lsst/ctrl/bps/panda/edgenode/cmd_line_decoder.py'

    executable = 'unset PYTHONPATH; source /cvmfs/sw.lsst.eu/linux-x86_64/lsst_distrib/w_2022_53/loadLSST.bash; pwd; ls -al; setup lsst_distrib;'

    executable += 'if [[ ! -z "${PANDA_AUTH_DIR}" ]] && [[ ! -z "${PANDA_AUTH_ORIGIN}" ]]; then export PANDA_AUTH_ID_TOKEN=$(cat $PANDA_AUTH_DIR); export PANDA_AUTH_VO=$PANDA_AUTH_ORIGIN; export IDDS_OIDC_TOKEN=$(cat $PANDA_AUTH_DIR); export IDDS_VO=$PANDA_AUTH_ORIGIN; export PANDA_AUTH=oidc; else unset PANDA_AUTH; export IDDS_AUTH_TYPE=x509_proxy; fi;'   # noqa E501

    # executable = 'python3 ${CTRL_BPS_PANDA_DIR}/python/lsst/ctrl/bps/panda/edgenode/cmd_line_decoder_build.py'
    # executable += 'wget https://storage.googleapis.com/drp-us-central1-containers/cmdline_builder.py;'
    executable += 'wget https://wguan-wisc.web.cern.ch/wguan-wisc/cmdline_builder.py;'
    executable += 'export IDDS_HOST=https://aipanda160.cern.ch:443/idds;'
    executable += 'python3 cmdline_builder.py '

    executable += output_filename + ' ' + './test_build.sh'

    work1 = DomaPanDAWork(executable=executable,
                          task_type='lsst_build',
                          primary_input_collection={'scope': 'pseudo_dataset', 'name': 'pseudo_input_collection#1'},
                          output_collections=[{'scope': 'pseudo_dataset', 'name': 'pseudo_output_collection#1'}],
                          log_collections=[], dependency_map=None,
                          task_name="build_task", task_queue=task_queue,
                          encode_command_line=True,
                          prodSourceLabel='managed',
                          task_log={"dataset": "PandaJob_#{pandaid}/",
                                    "destination": "local",
                                    "param_type": "log",
                                    "token": "local",
                                    "type": "template",
                                    "value": "log.tgz"},
                          task_cloud=task_cloud)

    pending_time = 12
    # pending_time = None
    workflow = Workflow(pending_time=pending_time)

    workflow.add_work(work1)
    workflow.name = 'test_workflow.idds.%s.test' % time.time()
    return workflow


if __name__ == '__main__':
    host = get_rest_host()
    workflow = setup_workflow()

    wm = ClientManager(host=host)
    # wm.set_original_user(user_name="wguandev")
    request_id = wm.submit_build(workflow, use_dataset_name=False)
    print(request_id)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020

import os
from traceback import format_exc

from flask import Blueprint, send_from_directory

from idds.common import exceptions
from idds.common.constants import HTTP_STATUS_CODE
from idds.common.utils import tar_zip_files, get_rest_cacher_dir
from idds.core import (transforms as core_transforms)
from idds.rest.v1.controller import IDDSController


class Logs(IDDSController):
    """  get(download) logs. """

    def get(self, workload_id, request_id):
        """ Get(download) logs.

        :param workload_id: the workload id.
        :param request_id: The id of the request.

        HTTP Success:
            200 OK
        HTTP Error:
            404 Not Found
            500 InternalError
        :returns: logs.
        """

        try:
            if workload_id == 'null':
                workload_id = None
            if request_id == 'null':
                request_id = None

            if workload_id is None and request_id is None:
                error = "One of workload_id and request_id should not be None or empty"
                return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

        try:
            transforms = core_transforms.get_transforms(request_id=request_id, workload_id=workload_id)
            workdirs = []
            for transform in transforms:
                work = transform['transform_metadata']['work']
                workdir = work.get_workdir()
                if workdir:
                    workdirs.append(workdir)
            if not workdirs:
                error = "No log files founded."
                return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

            cache_dir = get_rest_cacher_dir()
            if request_id and workload_id:
                output_filename = "request_%s.workload_%s.logs.tar.gz" % (request_id, workload_id)
            elif request_id:
                output_filename = "request_%s.logs.tar.gz" % (request_id)
            elif workload_id:
                output_filename = "workload_%s.logs.tar.gz" % (workload_id)
            else:
                output_filename = "%s.logs.tar.gz" % (os.path.basename(cache_dir))
            tar_zip_files(cache_dir, output_filename, workdirs)
            return send_from_directory(cache_dir, output_filename, as_attachment=True, mimetype='application/x-tgz')
        except exceptions.NoObject as error:
            return self.generate_http_response(HTTP_STATUS_CODE.NotFound, exc_cls=error.__class__.__name__, exc_msg=error)
        except exceptions.IDDSException as error:
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=error.__class__.__name__, exc_msg=error)
        except Exception as error:
            print(error)
            print(format_exc())
            return self.generate_http_response(HTTP_STATUS_CODE.InternalError, exc_cls=exceptions.CoreException.__name__, exc_msg=error)

    def post_test(self):
        import pprint
        pprint.pprint(self.get_request())
        pprint.pprint(self.get_request().endpoint)
        pprint.pprint(self.get_request().url_rule)


"""----------------------
   Web service url maps
----------------------"""


def get_blueprint():
    bp = Blueprint('logs', __name__)

    logs_view = Logs.as_view('logs')
    bp.add_url_rule('/logs/<workload_id>/<request_id>', view_func=logs_view, methods=['get', ])

    return bp



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2022

import traceback

from idds.common import exceptions
from idds.common.constants import ProcessingStatus, ProcessingLocking
from idds.common.utils import setup_logging, truncate_string
from idds.core import processings as core_processings
from idds.agents.common.eventbus.event import (EventType,
                                               NewProcessingEvent,
                                               SyncProcessingEvent,
                                               UpdateTransformEvent)

from .utils import handle_new_processing
from .poller import Poller

setup_logging(__name__)


class Submitter(Poller):
    """
    Submitter works to submit and running tasks to WFMS.
    """

    def __init__(self, num_threads=1, poll_period=10, retries=3, retrieve_bulk_size=2,
                 name='Submitter', message_bulk_size=1000, **kwargs):
        self.set_max_workers()
        num_threads = self.max_number_workers

        super(Submitter, self).__init__(num_threads=num_threads, name=name, **kwargs)

    def get_new_processings(self):
        """
        Get new processing
        """
        try:
            if not self.is_ok_to_run_more_processings():
                return []

            self.show_queue_size()

            processing_status = [ProcessingStatus.New]
            processings = core_processings.get_processings_by_status(status=processing_status, locking=True,
                                                                     not_lock=True,
                                                                     new_poll=True, only_return_id=True,
                                                                     bulk_size=self.retrieve_bulk_size)

            # self.logger.debug("Main thread get %s [new] processings to process" % len(processings))
            if processings:
                self.logger.info("Main thread get [new] processings to process: %s" % str(processings))

            for pr_id in processings:
                self.logger.info("NewProcessingEvent(processing_id: %s)" % pr_id)
                event = NewProcessingEvent(publisher_id=self.id, processing_id=pr_id)
                self.event_bus.send(event)

            return processings
        except exceptions.DatabaseException as ex:
            if 'ORA-00060' in str(ex):
                self.logger.warn("(cx_Oracle.DatabaseError) ORA-00060: deadlock detected while waiting for resource")
            else:
                # raise ex
                self.logger.error(ex)
                self.logger.error(traceback.format_exc())
        return []

    def handle_new_processing(self, processing):
        try:
            log_prefix = self.get_log_prefix(processing)

            # transform_id = processing['transform_id']
            # transform = core_transforms.get_transform(transform_id=transform_id)
            # work = transform['transform_metadata']['work']
            ret_new_processing = handle_new_processing(processing,
                                                       self.agent_attributes,
                                                       logger=self.logger,
                                                       log_prefix=log_prefix)
            status, processing, update_colls, new_contents, new_input_dependency_contents, msgs, errors = ret_new_processing

            if not status:
                raise exceptions.ProcessSubmitFailed(str(errors))

            parameters = {'status': ProcessingStatus.Submitting,
                          'substatus': ProcessingStatus.Submitting,
                          'locking': ProcessingLocking.Idle,
                          'processing_metadata': processing['processing_metadata']}
            parameters = self.load_poll_period(processing, parameters)

            proc = processing['processing_metadata']['processing']
            if proc.submitted_at:
                if not processing['submitted_at'] or processing['submitted_at'] < proc.submitted_at:
                    parameters['submitted_at'] = proc.submitted_at

            # if processing['processing_metadata'] and 'processing' in processing['processing_metadata']:
            if proc.workload_id:
                parameters['workload_id'] = proc.workload_id

            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': parameters}
            ret = {'update_processing': update_processing,
                   'update_collections': update_colls,
                   'update_contents': [],
                   'new_contents': new_contents,
                   'new_input_dependency_contents': new_input_dependency_contents,
                   'messages': msgs,
                   }
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())

            retries = processing['new_retries'] + 1
            if not processing['max_new_retries'] or retries < processing['max_new_retries']:
                pr_status = processing['status']
            else:
                pr_status = ProcessingStatus.Failed
            # increase poll period
            new_poll_period = int(processing['new_poll_period'].total_seconds() * self.poll_period_increase_rate)
            if new_poll_period > self.max_new_poll_period:
                new_poll_period = self.max_new_poll_period

            error = {'submit_err': {'msg': truncate_string('%s' % str(ex), length=200)}}
            parameters = {'status': pr_status,
                          'new_poll_period': new_poll_period,
                          'errors': processing['errors'] if processing['errors'] else {},
                          'new_retries': retries}
            parameters['errors'].update(error)

            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': parameters}
            ret = {'update_processing': update_processing,
                   'update_contents': []}
        return ret

    def process_new_processing(self, event):
        self.number_workers += 1
        try:
            if event:
                # pr_status = [ProcessingStatus.New]
                self.logger.info("process_new_processing, event: %s" % str(event))
                pr = self.get_processing(processing_id=event._processing_id, status=None, locking=True)
                if not pr:
                    self.logger.error("Cannot find processing for event: %s" % str(event))
                else:
                    log_pre = self.get_log_prefix(pr)
                    self.logger.info(log_pre + "process_new_processing")
                    ret = self.handle_new_processing(pr)
                    # self.logger.info(log_pre + "process_new_processing result: %s" % str(ret))

                    self.update_processing(ret, pr)

                    self.logger.info(log_pre + "UpdateTransformEvent(transform_id: %s)" % pr['transform_id'])
                    submit_event_content = {'event': 'submitted'}
                    event = UpdateTransformEvent(publisher_id=self.id, transform_id=pr['transform_id'], content=submit_event_content)
                    self.event_bus.send(event)

                    self.logger.info(log_pre + "SyncProcessingEvent(processing_id: %s)" % pr['processing_id'])
                    event = SyncProcessingEvent(publisher_id=self.id, processing_id=pr['processing_id'])
                    self.event_bus.send(event)
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
        self.number_workers -= 1

    def init_event_function_map(self):
        self.event_func_map = {
            EventType.NewProcessing: {
                'pre_check': self.is_ok_to_run_more_processings,
                'exec_func': self.process_new_processing
            }
        }

    def run(self):
        """
        Main run function.
        """
        try:
            self.logger.info("Starting main thread")

            self.load_plugins()
            self.init()

            self.add_default_tasks()

            self.init_event_function_map()

            task = self.create_task(task_func=self.get_new_processings, task_output_queue=None, task_args=tuple(), task_kwargs={}, delay_time=60, priority=1)
            self.add_task(task)

            self.execute()
        except KeyboardInterrupt:
            self.stop()


if __name__ == '__main__':
    agent = Submitter()
    agent()



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019


"""
operations related to Commands.
"""

import re
import datetime

import sqlalchemy
from sqlalchemy import or_
from sqlalchemy.exc import DatabaseError, IntegrityError

from idds.common import exceptions
from idds.common.constants import CommandLocation, CommandLocking
from idds.orm.base import models
from idds.orm.base.session import read_session, transactional_session


@transactional_session
def add_command(cmd_type, status, request_id, workload_id, transform_id,
                username=None, retries=0, processing_id=None,
                source=CommandLocation.Rest, destination=CommandLocation.Clerk,
                cmd_content=None, session=None):
    """
    Add a command to be submitted asynchronously to a command broker.

    :param cmd_type: The type of the cmd as a number, e.g., finished_stagein.
    :param status: The status about the command
    :param source: The source where the command is from.
    :param request_id: The request id.
    :param workload_id: The workload id.
    :param transform_id: The transform id.
    :param cmd_content: The command cmd_content as JSON.
    :param session: The database session.
    """

    try:
        cmd = models.Command(request_id=request_id, workload_id=workload_id,
                             transform_id=transform_id, cmd_type=cmd_type,
                             status=status, substatus=0, locking=0,
                             source=source, destination=destination,
                             username=username, retries=retries,
                             processing_id=processing_id,
                             cmd_content=cmd_content)

        cmd.save(session=session)
        cmd_id = cmd.cmd_id
        return cmd_id
    except TypeError as e:
        raise exceptions.DatabaseException('Invalid JSON for cmd_content: %s' % str(e))
    except DatabaseError as e:
        if re.match('.*ORA-12899.*', e.args[0]) \
           or re.match('.*1406.*', e.args[0]):
            raise exceptions.DatabaseException('Could not persist command, cmd_content too large: %s' % str(e))
        else:
            raise exceptions.DatabaseException('Could not persist command: %s' % str(e))


@transactional_session
def update_commands(commands, bulk_size=1000, session=None):
    try:
        session.bulk_update_mappings(models.Command, commands)
    except TypeError as e:
        raise exceptions.DatabaseException('Invalid JSON for cmd_content: %s' % str(e))
    except DatabaseError as e:
        if re.match('.*ORA-12899.*', e.args[0]) \
           or re.match('.*1406.*', e.args[0]):
            raise exceptions.DatabaseException('Could not persist command, cmd_content too large: %s' % str(e))
        else:
            raise exceptions.DatabaseException('Could not persist command: %s' % str(e))


@read_session
def retrieve_command(cmd_type=None, status=None, source=None,
                     destination=None, request_id=None, workload_id=None,
                     transform_id=None, processing_id=None, bulk_size=None, session=None):
    """
    Retrieve up to $bulk command.

    :param bulk: Number of command as an integer.
    :param cmd_type: Return only specified cmd_type.
    :param status: The status about the command
    :param source: The source where the command is from.
    :param session: The database session.

    :returns command: List of dictionaries
    """
    command = []
    try:
        query = session.query(models.Command)
        if request_id is not None:
            query = query.with_hint(models.Command, "INDEX(COMMANDS COMMANDS_TYPE_ST_IDX)", 'oracle')
        elif transform_id:
            query = query.with_hint(models.Command, "INDEX(COMMANDS COMMANDS_TYPE_ST_TF_IDX)", 'oracle')
        elif processing_id is not None:
            query = query.with_hint(models.Command, "INDEX(COMMANDS COMMANDS_TYPE_ST_PR_IDX)", 'oracle')
        else:
            query = query.with_hint(models.Command, "INDEX(COMMANDS COMMANDS_TYPE_ST_IDX)", 'oracle')

        if cmd_type is not None:
            query = query.filter_by(cmd_type=cmd_type)
        if status is not None:
            query = query.filter_by(status=status)
        if source is not None:
            query = query.filter_by(source=source)
        if destination is not None:
            query = query.filter_by(destination=destination)
        if request_id is not None:
            query = query.filter_by(request_id=request_id)
        if workload_id is not None:
            query = query.filter_by(workload_id=workload_id)
        if transform_id is not None:
            query = query.filter_by(transform_id=transform_id)
        if processing_id is not None:
            query = query.filter_by(processing_id=processing_id)

        if bulk_size:
            query = query.order_by(models.Command.created_at).limit(bulk_size)
        # query = query.with_for_update(nowait=True)

        tmp = query.all()
        if tmp:
            for t in tmp:
                command.append(t.to_dict())
        return command
    except IntegrityError as e:
        raise exceptions.DatabaseException(e.args)


@transactional_session
def delete_command(command, session=None):
    """
    Delete all command with the given IDs.

    :param command: The command to delete as a list of dictionaries.
    """
    command_condition = []
    for command in command:
        command_condition.append(models.Command.cmd_id == command['cmd_id'])

    try:
        if command_condition:
            session.query(models.Command).\
                with_hint(models.Command, "index(command COMMANDS_PK)", 'oracle').\
                filter(or_(*command_condition)).\
                delete(synchronize_session=False)
    except IntegrityError as e:
        raise exceptions.DatabaseException(e.args)


@transactional_session
def get_commands_by_status(status, locking=False, period=None, bulk_size=None, session=None):
    """
    Get commands

    :param status: Command status.
    :param locking: Whether only retrieves unlocked items.

    :param session: The database session in use.

    :returns: list of commands.
    """
    try:
        if status:
            if not isinstance(status, (list, tuple)):
                status = [status]
            if len(status) == 1:
                status = [status[0], status[0]]

        query = session.query(models.Command)
        if status:
            query = query.filter(models.Command.status.in_(status))

        if period:
            query = query.filter(models.Command.updated_at <= datetime.datetime.utcnow() - datetime.timedelta(seconds=period))

        if locking:
            query = query.filter(models.Command.locking == CommandLocking.Idle)
        # query = query.with_for_update(skip_locked=True)
        # query = query.order_by(asc(models.Command.updated_at))

        if bulk_size:
            query = query.limit(bulk_size)

        tmp = query.all()
        rets = []
        if tmp:
            rets = [t.to_dict() for t in tmp]
        return rets
    except sqlalchemy.orm.exc.NoResultFound as error:
        raise exceptions.NoObject('No commands attached with status (%s): %s' %
                                  (status, error))
    except Exception as error:
        raise error



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019


"""
Test Request.
"""

import unittest2 as unittest
from nose.tools import assert_equal

from idds.client.client import Client
from idds.common.constants import RequestStatus
from idds.common.utils import (check_database, has_config, setup_logging,
                               check_rest_host, get_rest_host, check_user_proxy)
from idds.orm.requests import (add_request, get_request, update_request,
                               delete_requests)
from idds.tests.common import get_request_properties

setup_logging(__name__)


class TestRequest(unittest.TestCase):

    @unittest.skipIf(not has_config(), "No config file")
    @unittest.skipIf(not check_database(), "Database is not defined")
    def test_create_and_check_for_request_orm(self):
        """ Request (ORM): Test the creation, query, and cancel of a Request """
        properties = get_request_properties()

        request_id = add_request(**properties)

        request = get_request(request_id=request_id)
        assert_equal(request_id, request['request_id'])

        for key in properties:
            if key in ['lifetime']:
                continue
            assert_equal(request[key], properties[key])

        request_id1 = add_request(**properties)
        delete_requests(request_id=request_id1)

        update_request(request_id, parameters={'status': RequestStatus.Failed})
        request = get_request(request_id=request_id)
        assert_equal(request['status'], RequestStatus.Failed)

        req = get_request(request_id=999999)
        assert_equal(req, None)

        delete_requests(request_id=request_id)

        req = get_request(request_id=request_id)
        assert_equal(req, None)

    @unittest.skipIf(not has_config(), "No config file")
    @unittest.skipIf(not check_user_proxy(), "No user proxy to access REST")
    @unittest.skipIf(not check_rest_host(), "REST host is not defined")
    def test_create_and_check_for_request_rest(self):
        """ Request (REST): Test the creation, query, and deletion of a Request """
        host = get_rest_host()

        properties = get_request_properties()

        client = Client(host=host)

        request_id = client.add_request(**properties)

        requests = client.get_requests(request_id=request_id)
        assert_equal(len(requests), 1)
        assert_equal(request_id, requests[0]['request_id'])

        for key in properties:
            if key in ['lifetime']:
                continue
            assert_equal(requests[0][key], properties[key])

        client.update_request(request_id, parameters={'status': RequestStatus.Failed})
        requests = client.get_requests(request_id=request_id)
        assert_equal(len(requests), 1)
        assert_equal(requests[0]['status'], RequestStatus.Failed)

        reqs = client.get_requests(request_id=999999)
        assert_equal(len(reqs), 0)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2022


import argparse
import logging


def config_api_host(conf_file_template="data/conf.js.template", conf_file='data/conf.js', hostname=None):
    with open(conf_file_template, 'r') as f:
        template = f.read()
    template = template.format(api_host_name=hostname)
    with open(conf_file, 'w') as f:
        f.write(template)


logging.getLogger().setLevel(logging.INFO)
parser = argparse.ArgumentParser(description="config iDDS monitor")
parser.add_argument('-s', '--source', default=None, help='Source config file path')
parser.add_argument('-d', '--destination', default=None, help='Destination file path')
parser.add_argument('--host', default=None, help='idds host name')
args = parser.parse_args()


config_api_host(conf_file_template=args.source, conf_file=args.destination, hostname=args.host)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020

from idds.common.constants import RequestType
from idds.common.utils import get_rest_host
from idds.core.requests import get_requests
from idds.workflowv2.workflow import Workflow

from idds.atlas.workflowv2.atlasstageinwork import ATLASStageinWork

from idds.client.clientmanager import ClientManager


def convert_req2reqv2(req):
    # v1: {'created_at': datetime.datetime(2020, 11, 3, 10, 9, 32), 'substatus': None, 'priority': 0, 'transform_tag': '2', 'requester': 'panda', 'request_metadata': {'workload_id': 23083304, 'rule_id': 'bef3da17f17c49ac97863bb9e96af672'}, 'name': 'valid1.361027.Pythia8EvtGen_A14NNPDF23LO_jetjet_JZ7W.simul.HITS.e5112_s3227_tid12560193_00', 'request_id': 3775, 'accessed_at': datetime.datetime(2020, 11, 3, 10, 9, 32), 'updated_at': datetime.datetime(2020, 11, 3, 10, 9, 32), 'locking': <RequestLocking.Idle: 0>, 'status': <RequestStatus.Cancelled: 9>, 'workload_id': 23083304, 'request_type': <RequestType.StageIn: 2>, 'errors': None, 'processing_metadata': None, 'scope': 'valid1', 'expired_at': datetime.datetime(2020, 12, 3, 10, 9, 32), 'next_poll_at': datetime.datetime(2020, 11, 3, 10, 9, 32)}  # noqa E501

    if req['request_type'] == RequestType.StageIn:
        request_metadata = req['request_metadata']
        work = ATLASStageinWork(executable=None, arguments=None, parameters=None, setup=None,
                                exec_type='local', sandbox=None,
                                primary_input_collection={'scope': req['scope'], 'name': req['name']},
                                other_input_collections=None,
                                output_collections={'scope': req['scope'], 'name': req['name'] + '.idds.stagein'},
                                log_collections=None,
                                logger=None,
                                max_waiting_time=request_metadata.get('max_waiting_time', 3600 * 7 * 24),
                                src_rse=request_metadata.get('src_rse', None),
                                dest_rse=request_metadata.get('dest_rse', None),
                                rule_id=request_metadata.get('rule_id', None))
    if req['request_type'] == RequestType.Workflow:
        pass
        ori_workflow = req['request_metadata']['workflow']
        ori_work = ori_workflow.works[ori_workflow.primary_initial_work]
        input_coll = ori_work.collections[ori_work.primary_input_collection]
        work = ATLASStageinWork(executable=None, arguments=None, parameters=None, setup=None,
                                exec_type='local', sandbox=None,
                                primary_input_collection={'scope': input_coll['scope'], 'name': input_coll['name']},
                                other_input_collections=None,
                                output_collections={'scope': input_coll['scope'], 'name': input_coll['name'] + '.idds.stagein'},
                                log_collections=None,
                                logger=None,
                                max_waiting_time=ori_work.max_waiting_time,
                                src_rse=ori_work.src_rse,
                                dest_rse=ori_work.dest_rse,
                                rule_id=ori_work.rule_id)

    workload_id = req['workload_id']
    if not workload_id and 'workload_id' in request_metadata:
        workload_id = request_metadata['workload_id']

    wf = Workflow()
    wf.set_workload_id(workload_id)
    wf.add_work(work)

    host = get_rest_host()
    wm = ClientManager(host=host)
    request_id = wm.submit(wf)
    # print(request_id)
    return request_id


reqs = get_requests()
print(len(reqs))
for req in reqs:
    # if req['request_id'] in [3743, 3755, 3769, 3775]:
    # if req['request_id'] in [3787, 3789, 3791]:
    if req['request_id'] in [11673, 17615, 19289, 19397]:
        # print(req)
        # print(req['request_metadata']['workflow'].to_dict())
        new_req_id = convert_req2reqv2(req)
        print("convert old request %s to new request %s" % (req['request_id'], new_req_id))
    # print(req['request_metadata']['workflow'].to_dict())
    pass




import json

from idds.core.messages import retrieve_messages

msgs = retrieve_messages()
for msg in msgs:
    if msg['msg_id'] in [323720]:
        # print(msg)
        msg_sent = json.dumps(msg['msg_content'])
        print(msg_sent)

        recover_msg = json.loads(msg_sent)
        if recover_msg['relation_type'] == 'output':
            # only parse messages from output contents
            points = recover_msg['files']
            for point in points:
                hp_loss = point['path']
                hp_loss = json.loads(hp_loss)
                hp, loss = hp_loss
                print(hp)
                print(loss)
                model_id, hp_point = hp
                print(model_id)
                print(hp_point)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2020


"""
operations related to Requests.
"""

import datetime

import sqlalchemy
from sqlalchemy.exc import DatabaseError, IntegrityError
from sqlalchemy.sql.expression import asc

from idds.common import exceptions
from idds.common.constants import CollectionType, CollectionStatus, CollectionLocking, CollectionRelationType
from idds.orm.base.session import read_session, transactional_session
from idds.orm.base import models


def create_collection(request_id, workload_id, scope, name, coll_type=CollectionType.Dataset, transform_id=None,
                      relation_type=CollectionRelationType.Input, bytes=0, status=CollectionStatus.New,
                      locking=CollectionLocking.Idle, total_files=0, new_files=0, processing_files=0,
                      processed_files=0, retries=0, expired_at=None,
                      coll_metadata=None):
    """
    Create a collection.

    :param scope: The scope of the request data.
    :param name: The name of the request data.
    :param coll_type: The type of dataset as dataset or container.
    :param request_id: The request id.
    :param workload_id: The workload id.
    :param transform_id: The transform id related to this collection.
    :param relation_type: The relation between this collection and its transform,
                          such as Input, Output, Log and so on.
    :param size: The size of the collection.
    :param status: The status.
    :param locking: The locking.
    :param total_files: Number of total files.
    :param retries: Number of retries.
    :param expired_at: The datetime when it expires.
    :param coll_metadata: The metadata as json.

    :returns: collection.
    """
    new_coll = models.Collection(request_id=request_id, workload_id=workload_id, scope=scope, name=name,
                                 coll_type=coll_type, transform_id=transform_id,
                                 relation_type=relation_type, bytes=bytes, status=status, locking=locking,
                                 total_files=total_files, new_files=new_files, processing_files=processing_files,
                                 processed_files=processed_files, retries=retries,
                                 expired_at=expired_at, coll_metadata=coll_metadata)
    return new_coll


@transactional_session
def add_collection(request_id, workload_id, scope, name, coll_type=CollectionType.Dataset, transform_id=None,
                   relation_type=CollectionRelationType.Input, bytes=0, status=CollectionStatus.New,
                   locking=CollectionLocking.Idle, total_files=0, new_files=0, processing_files=0,
                   processed_files=0, retries=0, expired_at=None,
                   coll_metadata=None, session=None):
    """
    Add a collection.

    :param scope: The scope of the request data.
    :param name: The name of the request data.
    :param coll_type: The type of dataset as dataset or container.
    :param request_id: The request id.
    :param workload_id: The workload id.
    :param transform_id: The transform id related to this collection.
    :param relation_type: The relation between this collection and its transform,
                          such as Input, Output, Log and so on.
    :param size: The size of the collection.
    :param status: The status.
    :param locking: The locking.
    :param total_files: Number of total files.
    :param retries: Number of retries.
    :param expired_at: The datetime when it expires.
    :param coll_metadata: The metadata as json.

    :raises DuplicatedObject: If a collection with the same name exists.
    :raises DatabaseException: If there is a database error.

    :returns: collection id.
    """
    try:
        new_coll = create_collection(request_id=request_id, workload_id=workload_id, scope=scope, name=name,
                                     coll_type=coll_type, transform_id=transform_id,
                                     relation_type=relation_type, bytes=bytes, status=status, locking=locking,
                                     total_files=total_files, new_files=new_files, retries=retries,
                                     processing_files=processing_files, processed_files=processed_files,
                                     expired_at=expired_at, coll_metadata=coll_metadata)
        new_coll.save(session=session)
        coll_id = new_coll.coll_id
        return coll_id
    except IntegrityError as error:
        raise exceptions.DuplicatedObject('Collection scope:name(%s:%s) with transform_id(%s) already exists!: %s' %
                                          (scope, name, transform_id, error))
    except DatabaseError as error:
        raise exceptions.DatabaseException(error)


@read_session
def get_collection_id(transform_id, relation_type, session=None):
    """
    Get collection id or raise a NoObject exception.

    :param transform_id: The transform id related to this collection.
    :param relation_type: The relation between this collection and its transform,
                          such as Input, Output, Log and so on.
    :param session: The database session in use.

    :raises NoObject: If no request is founded.

    :returns: Collection id.
    """

    try:
        query = session.query(models.Collection.coll_id)
        query = query.filter_by(transform_id=transform_id)
        query = query.filter(models.Collection.relation_type == relation_type)
        ret = query.first()
        if not ret:
            return None
        else:
            return ret[0]
    except sqlalchemy.orm.exc.NoResultFound as error:
        raise exceptions.NoObject('collection(transform_id: %s, relation_type: %s) cannot be found: %s' %
                                  (transform_id, relation_type, error))
    except Exception as error:
        raise error


@read_session
def get_collection(coll_id=None, transform_id=None, relation_type=None, to_json=False, session=None):
    """
    Get a collection or raise a NoObject exception.

    :param coll_id: The id of the collection.
    :param transform_id: The transform id related to this collection.
    :param relation_type: The relation between this collection and its transform,
                          such as Input, Output, Log and so on.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no request is founded.

    :returns: Collection.
    """

    try:
        if coll_id:
            query = session.query(models.Collection).filter_by(coll_id=coll_id)
        else:
            query = session.query(models.Collection).filter_by(transform_id=transform_id)\
                           .filter(models.Collection.relation_type == relation_type)

        ret = query.first()
        if not ret:
            return None
        else:
            if to_json:
                return ret.to_dict_json()
            else:
                return ret.to_dict()
    except sqlalchemy.orm.exc.NoResultFound as error:
        raise exceptions.NoObject('collection(coll_id: %s, transform_id: %s, relation_type: %s) cannot be found: %s' %
                                  (coll_id, transform_id, relation_type, error))
    except Exception as error:
        raise error


@read_session
def get_collection_ids_by_transform_id(transform_id=None, session=None):
    """
    Get collection ids by transform id or raise a NoObject exception.

    :param transform_id: The transform id related to this collection.
    :param session: The database session in use.

    :raises NoObject: If no collections are founded.

    :returns: list of Collection ids.
    """
    try:
        query = session.query(models.Collection.coll_id)\
                       .filter_by(transform_id=transform_id)
        ret = query.all()
        if not ret:
            return []
        else:
            items = []
            for t in ret:
                items.append(t[0])
            return items
    except sqlalchemy.orm.exc.NoResultFound as error:
        raise exceptions.NoObject('No collections with  transform_id(%s): %s' %
                                  (transform_id, error))
    except Exception as error:
        raise error


@read_session
def get_collections_by_status(status, relation_type=CollectionRelationType.Input, time_period=None,
                              locking=False, bulk_size=None, to_json=False, session=None):
    """
    Get collections by status, relation_type and time_period or raise a NoObject exception.

    :param status: The collection status.
    :param relation_type: The relation_type of the collection to the transform.
    :param time_period: time period in seconds since last update.
    :param locking: Wheter to retrieve unlocked files.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no collections are founded.

    :returns: list of Collections.
    """
    try:
        if not isinstance(status, (list, tuple)):
            status = [status]
        if len(status) == 1:
            status = [status[0], status[0]]

        query = session.query(models.Collection)\
                       .filter(models.Collection.status.in_(status))\
                       .filter(models.Collection.next_poll_at < datetime.datetime.utcnow())

        if relation_type is not None:
            query = query.filter(models.Collection.relation_type == relation_type)
        if time_period:
            query = query.filter(models.Collection.updated_at < datetime.datetime.utcnow() - datetime.timedelta(seconds=time_period))
        if locking:
            query = query.filter(models.Collection.locking == CollectionLocking.Idle)

        query = query.order_by(asc(models.Collection.updated_at))
        if bulk_size:
            query = query.limit(bulk_size)

        tmp = query.all()
        rets = []
        if tmp:
            for t in tmp:
                if to_json:
                    rets.append(t.to_dict_json())
                else:
                    rets.append(t.to_dict())
        return rets
    except sqlalchemy.orm.exc.NoResultFound as error:
        raise exceptions.NoObject('No collections with  status(%s), relation_type(%s), time_period(%s): %s' %
                                  (status, relation_type, time_period, error))
    except Exception as error:
        raise error


@read_session
def get_collections(scope=None, name=None, request_id=None, workload_id=None, transform_id=None,
                    relation_type=None, to_json=False, session=None):
    """
    Get collections by request id or raise a NoObject exception.

    :param scope: collection scope.
    :param name: collection name, can be wildcard.
    :param request_id: The request id.
    :param workload_id: The workload id.
    :param transform_id: list of transform id related to this collection.
    :param relation_type: The relation type between this collection and the transform: Input, Ouput and Log.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no collections are founded.

    :returns: list of Collections.
    """
    try:
        if transform_id and type(transform_id) not in (list, tuple):
            transform_id = [transform_id]

        query = session.query(models.Collection)
        if scope:
            query = query.filter(models.Collection.scope == scope)
        if name:
            query = query.filter(models.Collection.name.like(name.replace('*', '%')))
        if request_id:
            query = query.filter(models.Collection.request_id == request_id)
        if workload_id:
            query = query.filter(models.Collection.workload_id == workload_id)
        if transform_id:
            query = query.filter(models.Collection.transform_id.in_(transform_id))
        if relation_type is not None:
            query = query.filter(models.Collection.relation_type == relation_type)

        query = query.order_by(asc(models.Collection.updated_at))

        tmp = query.all()
        rets = []
        if tmp:
            for t in tmp:
                if to_json:
                    rets.append(t.to_dict_json())
                else:
                    rets.append(t.to_dict())
        return rets
    except sqlalchemy.orm.exc.NoResultFound as error:
        raise exceptions.NoObject('No collection with  scope(%s), name(%s), transform_id(%s): %s, relation_type: %s' %
                                  (scope, name, transform_id, relation_type, error))
    except Exception as error:
        raise error


@read_session
def get_collections_by_request_ids(request_ids, session=None):
    """"
    Get collections by a list of request ids.

    :param request_ids: list of request ids.

    :return collections: list of collections.
    """
    try:
        if request_ids and type(request_ids) not in (list, tuple):
            request_ids = [request_ids]

        query = session.query(models.Collection.coll_id,
                              models.Collection.request_id,
                              models.Collection.transform_id,
                              models.Collection.workload_id)
        if request_ids:
            query = query.filter(models.Collection.request_id.in_(request_ids))

        tmp = query.all()
        rets = []
        if tmp:
            for t in tmp:
                # rets.append(t.to_dict())
                t2 = dict(zip(t.keys(), t))
                rets.append(t2)
        return rets
    except Exception as error:
        raise error


@transactional_session
def update_collection(coll_id, parameters, session=None):
    """
    update a collection.

    :param coll_id: the collection id.
    :param parameters: A dictionary of parameters.
    :param session: The database session in use.

    :raises NoObject: If no request is founded.
    :raises DatabaseException: If there is a database error.

    """
    try:
        parameters['updated_at'] = datetime.datetime.utcnow()
        session.query(models.Collection).filter_by(coll_id=coll_id)\
               .update(parameters, synchronize_session=False)
    except sqlalchemy.orm.exc.NoResultFound as error:
        raise exceptions.NoObject('Collection %s cannot be found: %s' % (coll_id, error))


@transactional_session
def update_collections(parameters, session=None):
    """
    update collections.

    :param parameters: list of dictionary of parameters.
    :param session: The database session in use.

    :raises NoObject: If no content is founded.
    :raises DatabaseException: If there is a database error.

    """
    try:
        for parameter in parameters:
            parameter['updated_at'] = datetime.datetime.utcnow()

        session.bulk_update_mappings(models.Collection, parameters)
    except sqlalchemy.orm.exc.NoResultFound as error:
        raise exceptions.NoObject('Collection cannot be found: %s' % (error))


@transactional_session
def delete_collection(coll_id=None, session=None):
    """
    delete a collection.

    :param coll_id: The id of the collection.
    :param session: The database session in use.

    :raises NoObject: If no request is founded.
    :raises DatabaseException: If there is a database error.
    """
    try:
        session.query(models.Collection).filter_by(coll_id=coll_id).delete()
    except sqlalchemy.orm.exc.NoResultFound as error:
        raise exceptions.NoObject('Collection %s cannot be found: %s' % (coll_id, error))


@transactional_session
def clean_locking(time_period=3600, session=None):
    """
    Clearn locking which is older than time period.

    :param time_period in seconds
    """
    params = {'locking': 0}
    session.query(models.Collection).filter(models.Collection.locking == CollectionLocking.Locking)\
           .filter(models.Collection.updated_at < datetime.datetime.utcnow() - datetime.timedelta(seconds=time_period))\
           .update(params, synchronize_session=False)


@transactional_session
def clean_next_poll_at(status, session=None):
    """
    Clearn next_poll_at.

    :param status: status of the collection
    """
    if not isinstance(status, (list, tuple)):
        status = [status]
    if len(status) == 1:
        status = [status[0], status[0]]

    params = {'next_poll_at': datetime.datetime.utcnow()}
    session.query(models.Collection).filter(models.Collection.status.in_(status))\
           .update(params, synchronize_session=False)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020


"""
Test cacher.
"""

import sys

from idds.client.client import Client
from idds.common.utils import get_rest_host


if len(sys.argv) == 2:
    workload_id = sys.argv[1]
    request_id = None
elif len(sys.argv) == 3:
    workload_id = sys.argv[1]
    request_id = sys.argv[2]


host = get_rest_host()

client = Client(host=host)
filename = client.download_logs(workload_id=workload_id, request_id=request_id, dest_dir='/tmp')
print(filename)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019


"""
operations related to Transform.
"""

from idds.core import transforms


def add_transform(transform_type, transform_tag=None, priority=0, status=None, retries=0,
                  expired_at=None, transform_metadata=None, request_id=None):
    """
    Add a transform.

    :param transform_type: Transform type.
    :param transform_tag: Transform tag.
    :param priority: priority.
    :param status: Transform status.
    :param retries: The number of retries.
    :param expired_at: The datetime when it expires.
    :param transform_metadata: The metadata as json.

    :returns: content id.
    """
    kwargs = {'transform_type': transform_type, 'transform_tag': transform_tag, 'priority': priority,
              'status': status, 'retries': retries, 'expired_at': expired_at,
              'transform_metadata': transform_metadata, 'request_id': request_id}
    return transforms.add_transform(**kwargs)


def get_transform(transform_id):
    """
    Get transform.

    :param transform_id: Transform id.

    :returns: Transform.
    """
    return transforms.get_transform(transform_id)


def update_transform(transform_id, parameters):
    """
    update a transform.

    :param transform_id: the transform id.
    :param parameters: A dictionary of parameters.

    """
    return transforms.update_transform(transform_id, parameters)


def delete_transform(transform_id=None):
    """
    delete a transform.

    :param transform_id: The id of the transform.
    """
    return transforms.delete_transform(transform_id)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019


"""
Test Request.
"""

import unittest2 as unittest
from nose.tools import assert_equal

from idds.common.utils import check_database, has_config, setup_logging
from idds.common.constants import ProcessingStatus
from idds.orm.transforms import add_transform, delete_transform
from idds.orm.processings import (add_processing, update_processing,
                                  get_processing, delete_processing)
from idds.tests.common import get_transform_properties, get_processing_properties

setup_logging(__name__)


class TestTransformProcessing(unittest.TestCase):

    @unittest.skipIf(not has_config(), "No config file")
    @unittest.skipIf(not check_database(), "Database is not defined")
    def test_create_and_check_for_transform_processing_orm(self):
        """ Transform/Processing (ORM): Test the creation, query, and cancel of a Transform/Processing """
        trans_properties = get_transform_properties()

        proc_properties = get_processing_properties()

        trans_id = add_transform(**trans_properties)

        proc_properties['transform_id'] = trans_id
        processing_id = add_processing(**proc_properties)
        processing = get_processing(processing_id=processing_id)
        for key in proc_properties:
            assert_equal(processing[key], proc_properties[key])

        update_processing(processing_id, {'status': ProcessingStatus.Failed})
        processing = get_processing(processing_id=processing_id)
        assert_equal(processing['status'], ProcessingStatus.Failed)

        delete_processing(processing_id)

        processing = get_processing(processing_id=processing_id)
        assert_equal(processing, None)

        delete_transform(trans_id)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2021


"""
Test client.
"""
import json
import re
import time
# import traceback

try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote

from pandaclient import Client

# from idds.client.client import Client
from idds.client.clientmanager import ClientManager
# from idds.common.constants import RequestType, RequestStatus
from idds.common.utils import get_rest_host, run_command
# from idds.common.utils import json_dumps
# from idds.tests.common import get_example_real_tape_stagein_request
# from idds.tests.common import get_example_prodsys2_tape_stagein_request

# from idds.workflowv2.work import Work, Parameter, WorkStatus
from idds.workflowv2.workflow import Condition, Workflow
# from idds.workflowv2.workflow import Workflow
# from idds.atlas.workflowv2.atlasstageinwork import ATLASStageinWork
from idds.atlas.workflowv2.atlaspandawork import ATLASPandaWork
from idds.atlas.workflowv2.atlasactuatorwork import ATLASActuatorWork


def get_task_id(output, error):
    m = re.search('jediTaskID=(\d+)', output + error)  # noqa W605
    task_id = int(m.group(1))
    return task_id


def submit_processing_task():
    outDS = "user.wguan.altest%s" % str(int(time.time()))
    cmd = "cd /afs/cern.ch/user/w/wguan/workdisk/iDDS/main/lib/idds/tests/activelearning_test_codes; prun --exec 'python simplescript.py 0.5 0.5 200 output.json' --outDS %s  --outputs output.json --nJobs=10" % outDS
    status, output, error = run_command(cmd)
    """
    print("status:")
    print(status)
    print("output:")
    print(output)
    print("error:")
    print(error)

    status:
    0
    output:

    error:
    INFO : gathering files under /afs/cern.ch/user/w/wguan/workdisk/iDDS/main/lib/idds/tests/activelearning_test_codes
    INFO : upload source files
    INFO : submit user.wguan.altest1234/
    INFO : succeeded. new jediTaskID=23752996
    """
    if status == 0:
        task_id = get_task_id(output, error)
        return task_id
    else:
        raise Exception(output + error)


def get_panda_task_paramsmap(panda_task_id):
    status, task_param_map = Client.getTaskParamsMap(panda_task_id)
    if status == 0:
        task_param_map = json.loads(task_param_map)
        return task_param_map
    return None


def define_panda_task_paramsmap():
    # here is using a fake method by submitting a panda task.
    # Users should define the task params map by themselves.

    # (0, '{"buildSpec": {"jobParameters": "-i ${IN} -o ${OUT} --sourceURL ${SURL} -r . ", "archiveName": "sources.0ca6a2fb-4ad0-42d0-979d-aa7c284f1ff7.tar.gz", "prodSourceLabel": "panda"}, "sourceURL": "https://aipanda048.cern.ch:25443", "cliParams": "prun --exec \\"python simplescript.py 0.5 0.5 200 output.json\\" --outDS user.wguan.altest1234 --outputs output.json --nJobs=10", "site": null, "vo": "atlas", "respectSplitRule": true, "osInfo": "Linux-3.10.0-1127.19.1.el7.x86_64-x86_64-with-centos-7.9.2009-Core", "log": {"type": "template", "param_type": "log", "container": "user.wguan.altest1234.log/", "value": "user.wguan.altest1234.log.$JEDITASKID.${SN}.log.tgz", "dataset": "user.wguan.altest1234.log/"}, "transUses": "", "excludedSite": [], "nMaxFilesPerJob": 200, "uniqueTaskName": true, "noInput": true, "taskName": "user.wguan.altest1234/", "transHome": null, "includedSite": null, "nEvents": 10, "nEventsPerJob": 1, "jobParameters": [{"type": "constant", "value": "-j \\"\\" --sourceURL ${SURL}"}, {"type": "constant", "value": "-r ."}, {"padding": false, "type": "constant", "value": "-p \\""}, {"padding": false, "type": "constant", "value": "python%20simplescript.py%200.5%200.5%20200%20output.json"}, {"type": "constant", "value": "\\""}, {"type": "constant", "value": "-l ${LIB}"}, {"container": "user.wguan.altest1234_output.json/", "value": "user.wguan.$JEDITASKID._${SN/P}.output.json", "dataset": "user.wguan.altest1234_output.json/", "param_type": "output", "hidden": true, "type": "template"}, {"type": "constant", "value": "-o \\"{\'output.json\': \'user.wguan.$JEDITASKID._${SN/P}.output.json\'}\\""}], "prodSourceLabel": "user", "processingType": "panda-client-1.4.47-jedi-run", "architecture": "@centos7", "userName": "Wen Guan", "taskType": "anal", "taskPriority": 1000, "countryGroup": "us"}')  # noqa E501

    task_id = submit_processing_task()
    task_param_map = get_panda_task_paramsmap(task_id)
    cmd_to_arguments = {'arguments': 'python simplescript.py 0.5 0.5 200',
                        'parameters': 'python simplescript.py {m1} {m2} {nevents}'}

    # update the cliParams to have undefined parameters, these parameters {m1}, {m2}, {nevents} will be the outputs of learning script.
    task_param_map['cliParams'] = task_param_map['cliParams'].replace(cmd_to_arguments['arguments'], cmd_to_arguments['parameters'])
    jobParameters = task_param_map['jobParameters']
    for p in jobParameters:
        if 'value' in p:
            p['value'] = p['value'].replace(quote(cmd_to_arguments['arguments']), quote(cmd_to_arguments['parameters']))
    return task_param_map


def test_panda_work():
    task_param_map = define_panda_task_paramsmap()
    work = ATLASPandaWork(panda_task_paramsmap=task_param_map)
    work.initialize_work()
    print(work.__class__.__name__)
    print('sandbox: %s' % work.sandbox)
    print('output_collections: %s' % str(work.get_output_collections()))

    print("new work")
    test_work = work.generate_work_from_template()
    test_work.set_parameters({'m1': 0.5, 'm2': 0.5, 'nevents': 100})

    test_work.initialize_work()
    # print(json_dumps(test_work, sort_keys=True, indent=4))
    # print('output_collections: %s' % str(test_work.get_output_collections()))
    # print(json_dumps(test_work, sort_keys=True, indent=4))

    # from pandaclient import Client
    # Client.getJediTaskDetails(taskDict,fullFlag,withTaskInfo,verbose=False)
    # ret = Client.getJediTaskDetails({'jediTaskID': panda_task_id},False,True)
    # print(ret)


def get_workflow():
    task_param_map = define_panda_task_paramsmap()
    work = ATLASPandaWork(panda_task_paramsmap=task_param_map)

    # it's needed to parse the panda task parameter information, for example output dataset name, for the next task.
    # if the information is not needed, you don't need to run it manually. iDDS will call it interally to parse the information.
    work.initialize_work()

    work_output_coll = work.get_output_collections()[0]

    input_coll = {'scope': work_output_coll['scope'],
                  'name': work_output_coll['name'],
                  'coll_metadata': {'force_close': True}}
    output_coll = {'scope': work_output_coll['scope'],
                   'name': work_output_coll['name'] + "." + str(int(time.time()))}

    # acutator = ATLASActuatorWork(executable='python', arguments='merge.py {output_json} {events} {dataset}/{filename}',
    acutator = ATLASActuatorWork(executable='python', arguments='merge.py {output_json} {events} {dataset}',
                                 parameters={'output_json': 'merge.json',
                                             'events': 200,
                                             'dataset': '{scope}:{name}'.format(**input_coll),
                                             'filename': 'output*.json'},
                                 sandbox=work.sandbox, primary_input_collection=input_coll,
                                 output_collections=output_coll, output_json='merge.json')
    wf = Workflow()
    # because the two tasks are in a loop. It's good to set which one to start.
    wf.add_work(work)
    wf.add_work(acutator)
    cond = Condition(work.is_finished, current_work=work, true_work=acutator, false_work=None)
    wf.add_condition(cond)
    cond1 = Condition(acutator.generate_new_task, current_work=acutator, true_work=work, false_work=None)
    wf.add_condition(cond1)

    # because the two works are in a loop, they are not independent. This call is needed to tell which one to start.
    # otherwise idds will use the first one to start.
    wf.add_initial_works(work)

    # work.set_workflow(wf)
    return wf


if __name__ == '__main__':
    host = get_rest_host()
    # panda_task_id = submit_processing_task()
    # panda_task_id = 23752996
    # panda_task_id = 23810059
    # panda_task_id = 23818866
    # print(panda_task_id)
    test_panda_work()
    workflow = get_workflow()
    wm = ClientManager(host=host)
    request_id = wm.submit(workflow)
    print(request_id)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2023

import time
import traceback
try:
    # python 3
    from queue import Queue
except ImportError:
    # Python 2
    from Queue import Queue

from idds.common.constants import Sections
from idds.common.exceptions import AgentPluginError, IDDSException
from idds.common.utils import setup_logging, get_logger
from idds.common.utils import json_dumps
from idds.core import messages as core_messages, catalog as core_catalog
from idds.agents.common.baseagent import BaseAgent
# from idds.agents.common.eventbus.event import TerminatedProcessingEvent
from idds.agents.common.eventbus.event import MsgTriggerProcessingEvent

from .utils import handle_messages_processing

setup_logging(__name__)


class Receiver(BaseAgent):
    """
    Receiver works to receive workload management messages to update task/job status.
    """

    def __init__(self, num_threads=1, bulk_message_delay=30, bulk_message_size=2000,
                 random_delay=None, update_processing_interval=300, **kwargs):
        super(Receiver, self).__init__(num_threads=num_threads, name='Receiver', **kwargs)
        self.config_section = Sections.Carrier
        self.bulk_message_delay = int(bulk_message_delay)
        self.bulk_message_size = int(bulk_message_size)
        self.message_queue = Queue()
        self.logger = get_logger(self.__class__.__name__)
        self.update_processing_interval = update_processing_interval
        if self.update_processing_interval:
            self.update_processing_interval = int(self.update_processing_interval)
        else:
            self.update_processing_interval = 300

    def __del__(self):
        self.stop_receiver()

    def start_receiver(self):
        if 'receiver' not in self.plugins:
            raise AgentPluginError('Plugin receiver is required')
        self.receiver = self.plugins['receiver']

        self.logger.info("Starting receiver: %s" % self.receiver)
        self.receiver.set_output_queue(self.message_queue)
        self.setup_logger(self.logger)
        self.receiver.start()

    def stop_receiver(self):
        if hasattr(self, 'receiver') and self.receiver:
            self.logger.info("Stopping receiver: %s" % self.receiver)
            self.receiver.stop()

    def get_output_messages(self):
        msgs = []
        try:
            while not self.message_queue.empty():
                msg = self.message_queue.get(False)
                if msg:
                    self.logger.debug("Received message: %s" % str(msg))
                    msgs.append(msg)
        except Exception as error:
            self.logger.error("Failed to get output messages: %s, %s" % (error, traceback.format_exc()))
        return msgs

    def run(self):
        """
        Main run function.
        """
        try:
            self.logger.info("Starting main thread")
            self.load_plugins()

            self.start_receiver()

            self.add_health_message_task()

            log_prefix = "<Message>"

            while not self.graceful_stop.is_set():
                try:
                    time_start = time.time()
                    output_messages = self.get_output_messages()
                    update_processings, terminated_processings, update_contents, msgs = handle_messages_processing(output_messages,
                                                                                                                   logger=self.logger,
                                                                                                                   log_prefix=log_prefix,
                                                                                                                   update_processing_interval=self.update_processing_interval)

                    if msgs:
                        # self.logger.debug(log_prefix + "adding messages[:3]: %s" % json_dumps(msgs[:3]))
                        core_messages.add_messages(msgs, bulk_size=self.bulk_message_size)

                    if update_contents:
                        self.logger.info(log_prefix + "update_contents[:3]: %s" % json_dumps(update_contents[:3]))
                        core_catalog.update_contents(update_contents)

                    for pr_id in update_processings:
                        # self.logger.info(log_prefix + "TerminatedProcessingEvent(processing_id: %s)" % pr_id)
                        # event = TerminatedProcessingEvent(publisher_id=self.id, processing_id=pr_id)
                        self.logger.info(log_prefix + "MsgTriggerProcessingEvent(processing_id: %s)" % pr_id)
                        event = MsgTriggerProcessingEvent(publisher_id=self.id, processing_id=pr_id)
                        self.event_bus.send(event)

                    for pr_id in terminated_processings:
                        self.logger.info(log_prefix + "MsgTriggerProcessingEvent(processing_id: %s)" % pr_id)
                        event = MsgTriggerProcessingEvent(publisher_id=self.id, processing_id=pr_id, content={'Terminated': True, 'source': 'Receiver'})
                        self.event_bus.send(event)

                    time_delay = self.bulk_message_delay - (time.time() - time_start)
                    time_delay = self.bulk_message_delay
                    if time_delay > 0:
                        time.sleep(time_delay)
                except IDDSException as error:
                    self.logger.error("Main thread IDDSException: %s" % str(error))
                except Exception as error:
                    self.logger.critical("Main thread exception: %s\n%s" % (str(error), traceback.format_exc()))
        except KeyboardInterrupt:
            self.stop()

    def stop(self):
        super(Receiver, self).stop()
        self.stop_receiver()


if __name__ == '__main__':
    agent = Receiver()
    agent()



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2022 - 2023


import argparse
import logging

import os
import re
import json

import configparser


is_unicode_defined = True
try:
    _ = unicode('test')
except NameError:
    is_unicode_defined = False


def is_string(value):
    if is_unicode_defined and type(value) in [str, unicode] or type(value) in [str]:   # noqa F821
        return True
    else:
        return False


def as_parse_env(dct):
    for key in dct:
        value = dct[key]
        if is_string(value) and '$' in value:
            env_matches = re.findall('\$\{*([^\}]+)\}*', value)     # noqa W605
            for env_name in env_matches:
                if env_name not in os.environ:
                    print("Error: %s is defined in configmap but is not defined in environments" % env_name)
                else:
                    env_name1 = r'${%s}' % env_name
                    env_name2 = r'$%s' % env_name
                    value = value.replace(env_name1, os.environ.get(env_name)).replace(env_name2, os.environ.get(env_name))
        dct[key] = value
    return dct


def convert_section_json(data_conf):
    for section in data_conf:
        for item in data_conf[section]:
            if type(data_conf[section][item]) in [list, tuple, dict]:
                data_conf[section][item] = json.dumps(data_conf[section][item])
    return data_conf


def merge_configs(source_file_path, dest_file_path):
    """
    Merge  configuration file.
    """

    if source_file_path and dest_file_path:
        if os.path.exists(source_file_path) and os.path.exists(dest_file_path):
            with open(source_file_path, 'r') as f:
                data = json.load(f, object_hook=as_parse_env)
                if dest_file_path in data:
                    data_conf = data[dest_file_path]
                    data_conf = convert_section_json(data_conf)
                    parser = configparser.ConfigParser()
                    parser.read(dest_file_path)
                    parser.read_dict(data_conf)
                    with open(dest_file_path, 'w') as dest_file:
                        parser.write(dest_file)


def create_oidc_token():
    if 'PANDA_AUTH_ID_TOKEN' in os.environ:
        config_root = os.environ.get('PANDA_CONFIG_ROOT', '/tmp')
        token_file = os.path.join(config_root, '.token')
        token = {"id_token": os.environ.get('PANDA_AUTH_ID_TOKEN', ""),
                 "token_type": "Bearer"}

        with open(token_file, 'w') as f:
            f.write(json.dumps(token))


logging.getLogger().setLevel(logging.INFO)
parser = argparse.ArgumentParser(description="Merge configuration file from configmap")
parser.add_argument('-s', '--source', default=None, help='Source config file path (in .json format)')
parser.add_argument('-d', '--destination', default=None, help='Destination file path')
parser.add_argument('-c', '--create_oidc_token', default=False, action='store_true', help='Create the oidc token based on environment')
args = parser.parse_args()

merge_configs(args.source, args.destination)
if args.create_oidc_token:
    create_oidc_token()



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2022


"""
operations related to Transform.
"""

import datetime
import logging

# from idds.common import exceptions

from idds.common.constants import (TransformStatus, ContentRelationType, ContentStatus,
                                   TransformLocking, CollectionRelationType)
from idds.orm.base.session import read_session, transactional_session
from idds.orm import (transforms as orm_transforms,
                      collections as orm_collections,
                      contents as orm_contents,
                      messages as orm_messages,
                      processings as orm_processings)


@transactional_session
def add_transform(request_id, workload_id, transform_type, transform_tag=None, priority=0, name=None,
                  status=TransformStatus.New, substatus=TransformStatus.New, locking=TransformLocking.Idle,
                  new_poll_period=1, update_poll_period=10, retries=0, expired_at=None, transform_metadata=None,
                  new_retries=0, update_retries=0, max_new_retries=3, max_update_retries=0,
                  workprogress_id=None, session=None):
    """
    Add a transform.

    :param request_id: The request id.
    :param workload_id: The workload id.
    :param transform_type: Transform type.
    :param transform_tag: Transform tag.
    :param priority: priority.
    :param status: Transform status.
    :param locking: Transform locking.
    :param retries: The number of retries.
    :param expired_at: The datetime when it expires.
    :param transform_metadata: The metadata as json.

    :raises DuplicatedObject: If a transform with the same name exists.
    :raises DatabaseException: If there is a database error.

    :returns: transform id.
    """
    transform_id = orm_transforms.add_transform(request_id=request_id, workload_id=workload_id,
                                                transform_type=transform_type, transform_tag=transform_tag,
                                                priority=priority, status=status, substatus=substatus,
                                                locking=locking, retries=retries, name=name,
                                                new_poll_period=new_poll_period,
                                                update_poll_period=update_poll_period,
                                                new_retries=new_retries, update_retries=update_retries,
                                                max_new_retries=max_new_retries,
                                                max_update_retries=max_update_retries,
                                                expired_at=expired_at, transform_metadata=transform_metadata,
                                                workprogress_id=workprogress_id, session=session)
    return transform_id


@read_session
def get_transform(transform_id, to_json=False, session=None):
    """
    Get transform or raise a NoObject exception.

    :param transform_id: Transform id.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no transform is founded.

    :returns: Transform.
    """
    return orm_transforms.get_transform(transform_id=transform_id, to_json=to_json, session=session)


@transactional_session
def get_transform_by_id_status(transform_id, status=None, locking=False, session=None):
    tf = orm_transforms.get_transform_by_id_status(transform_id=transform_id, status=status, locking=locking, session=session)
    if tf is not None and locking:
        parameters = {}
        parameters['locking'] = TransformLocking.Locking
        parameters['updated_at'] = datetime.datetime.utcnow()
        orm_transforms.update_transform(transform_id=tf['transform_id'], parameters=parameters, session=session)
    return tf


@read_session
def get_transforms_with_input_collection(transform_type, transform_tag, coll_scope, coll_name, to_json=False, session=None):
    """
    Get transform or raise a NoObject exception.

    :param transform_type: Transform type.
    :param transform_tag: Transform tag.
    :param coll_scope: The collection scope.
    :param coll_name: The collection name.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no transform is founded.

    :returns: Transforms.
    """
    return orm_transforms.get_transforms_with_input_collection(transform_type, transform_tag, coll_scope,
                                                               coll_name, to_json=to_json, session=session)


@read_session
def get_transform_ids(workprogress_id, request_id=None, workload_id=None, transform_id=None, session=None):
    """
    Get transform ids or raise a NoObject exception.

    :param workprogress_id: Workprogress id.
    :param session: The database session in use.

    :raises NoObject: If no transform is founded.

    :returns: list of transform ids.
    """
    return orm_transforms.get_transform_ids(workprogress_id=workprogress_id, request_id=request_id,
                                            workload_id=workload_id, transform_id=transform_id, session=session)


@read_session
def get_transforms(request_id=None, workload_id=None, transform_id=None, to_json=False, session=None):
    """
    Get transforms or raise a NoObject exception.

    :param workprogress_id: Workprogress id.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no transform is founded.

    :returns: list of transform.
    """
    return orm_transforms.get_transforms(request_id=request_id,
                                         workload_id=workload_id,
                                         transform_id=transform_id,
                                         to_json=to_json, session=session)


@transactional_session
def get_transforms_by_status(status, period=None, locking=False, bulk_size=None, to_json=False, by_substatus=False,
                             new_poll=False, update_poll=False, only_return_id=False,
                             not_lock=False, next_poll_at=None, session=None):
    """
    Get transforms or raise a NoObject exception.

    :param status: Transform status or list of transform status.
    :param session: The database session in use.
    :param locking: Whether to lock retrieved items.
    :param to_json: return json format.

    :raises NoObject: If no transform is founded.

    :returns: list of transform.
    """
    if locking:
        if not only_return_id and bulk_size:
            # order by cannot work together with locking. So first select 2 * bulk_size without locking with order by.
            # then select with locking.
            tf_ids = orm_transforms.get_transforms_by_status(status=status, period=period, locking=locking,
                                                             bulk_size=bulk_size * 2, locking_for_update=False,
                                                             to_json=False, only_return_id=True,
                                                             new_poll=new_poll, update_poll=update_poll,
                                                             by_substatus=by_substatus, session=session)
            if tf_ids:
                transform2s = orm_transforms.get_transforms_by_status(status=status, period=period, locking=locking,
                                                                      bulk_size=None, locking_for_update=False,
                                                                      to_json=to_json, transform_ids=tf_ids,
                                                                      new_poll=new_poll, update_poll=update_poll,
                                                                      by_substatus=by_substatus, session=session)
                if transform2s:
                    # reqs = req2s[:bulk_size]
                    # order requests
                    transforms = []
                    for tf_id in tf_ids:
                        if len(transforms) >= bulk_size:
                            break
                        for tf in transform2s:
                            if tf['transform_id'] == tf_id:
                                transforms.append(tf)
                                break
                    # transforms = transforms[:bulk_size]
                else:
                    transforms = []
            else:
                transforms = []
        else:
            transforms = orm_transforms.get_transforms_by_status(status=status, period=period, locking=locking,
                                                                 locking_for_update=False,
                                                                 bulk_size=bulk_size, to_json=to_json,
                                                                 new_poll=new_poll, update_poll=update_poll,
                                                                 only_return_id=only_return_id,
                                                                 by_substatus=by_substatus, session=session)

        parameters = {}
        if not not_lock:
            parameters['locking'] = TransformLocking.Locking
        if next_poll_at:
            parameters['next_poll_at'] = next_poll_at
        parameters['updated_at'] = datetime.datetime.utcnow()
        if parameters:
            for transform in transforms:
                if type(transform) in [dict]:
                    orm_transforms.update_transform(transform_id=transform['transform_id'], parameters=parameters, session=session)
                else:
                    orm_transforms.update_transform(transform_id=transform, parameters=parameters, session=session)
    else:
        transforms = orm_transforms.get_transforms_by_status(status=status, period=period, locking=locking,
                                                             bulk_size=bulk_size, to_json=to_json,
                                                             new_poll=new_poll, update_poll=update_poll,
                                                             only_return_id=only_return_id,
                                                             by_substatus=by_substatus, session=session)
    return transforms


@transactional_session
def update_transform(transform_id, parameters, session=None):
    """
    update a transform.

    :param transform_id: the transform id.
    :param parameters: A dictionary of parameters.
    :param session: The database session in use.

    :raises NoObject: If no content is founded.
    :raises DatabaseException: If there is a database error.

    """
    orm_transforms.update_transform(transform_id=transform_id, parameters=parameters, session=session)


@transactional_session
def add_transform_outputs(transform, transform_parameters, input_collections=None, output_collections=None, log_collections=None,
                          update_input_collections=None, update_output_collections=None, update_log_collections=None,
                          new_contents=None, update_contents=None, new_processing=None, update_processing=None,
                          messages=None, update_messages=None, message_bulk_size=10000, session=None):
    """
    For input contents, add corresponding output contents.

    :param transform: the transform.
    :param input_collections: The new input collections.
    :param output_collections: The new output collections.
    :param log_collections: The new log collections.
    :param update_input_collections: The updated input collections.
    :param update_output_collections: The updated output collections.
    :param update_log_collections: The updated log collections.
    :param new_contents: The new contents.
    :param update_contents: The updated contents.
    :param new_processing: The new processing.
    :param messages: Messages.
    :param message_bulk_size: The message bulk size.
    :param session: The database session in use.

    :raises DatabaseException: If there is a database error.
    """
    work = transform['transform_metadata']['work']

    new_pr_ids, update_pr_ids = [], []

    if input_collections:
        for coll in input_collections:
            collection = coll['collection']
            del coll['collection']
            coll_id = orm_collections.add_collection(**coll, session=session)
            # work.set_collection_id(coll, coll_id)
            collection.coll_id = coll_id
    if output_collections:
        for coll in output_collections:
            collection = coll['collection']
            del coll['collection']
            coll_id = orm_collections.add_collection(**coll, session=session)
            # work.set_collection_id(coll, coll_id)
            collection.coll_id = coll_id
    if log_collections:
        for coll in log_collections:
            collection = coll['collection']
            del coll['collection']
            coll_id = orm_collections.add_collection(**coll, session=session)
            # work.set_collection_id(coll, coll_id)
            collection.coll_id = coll_id

    if update_input_collections:
        update_input_colls = [coll.collection for coll in update_input_collections]
        orm_collections.update_collections(update_input_colls, session=session)
    if update_output_collections:
        update_output_colls = [coll.collection for coll in update_output_collections]
        orm_collections.update_collections(update_output_colls, session=session)
    if update_log_collections:
        update_log_colls = [coll.collection for coll in update_log_collections]
        orm_collections.update_collections(update_log_colls, session=session)

    if new_contents:
        orm_contents.add_contents(new_contents, session=session)
    if update_contents:
        orm_contents.update_contents(update_contents, session=session)

    processing_id = None
    if new_processing:
        # print(new_processing)
        processing_id = orm_processings.add_processing(**new_processing, session=session)
        new_pr_ids.append(processing_id)
    if update_processing:
        for proc_id in update_processing:
            orm_processings.update_processing(processing_id=proc_id, parameters=update_processing[proc_id], session=session)
            update_pr_ids.append(proc_id)

    if messages:
        if not type(messages) in [list, tuple]:
            messages = [messages]
        # for message in messages:
        #     orm_messages.add_message(msg_type=message['msg_type'],
        #                              status=message['status'],
        #                              source=message['source'],
        #                              request_id=message['request_id'],
        #                              workload_id=message['workload_id'],
        #                              transform_id=message['transform_id'],
        #                              num_contents=message['num_contents'],
        #                              msg_content=message['msg_content'],
        #                              bulk_size=message_bulk_size,
        #                              session=session)
        logging.debug("message_bulk_size: %s" % str(message_bulk_size))
        orm_messages.add_messages(messages, bulk_size=message_bulk_size, session=session)
    if update_messages:
        orm_messages.update_messages(update_messages, bulk_size=message_bulk_size, session=session)

    if transform:
        if processing_id:
            # work.set_processing_id(new_processing, processing_id)
            work.set_processing_id(new_processing['processing_metadata']['processing'], processing_id)
        work.refresh_work()
        orm_transforms.update_transform(transform_id=transform['transform_id'],
                                        parameters=transform_parameters,
                                        session=session)
    return new_pr_ids, update_pr_ids


@transactional_session
def delete_transform(transform_id=None, session=None):
    """
    delete a transform.

    :param transform_id: The id of the transform.
    :param session: The database session in use.

    :raises NoObject: If no content is founded.
    :raises DatabaseException: If there is a database error.
    """
    orm_transforms.delete_transform(transform_id=transform_id, session=session)


@transactional_session
def clean_locking(time_period=3600, session=None):
    """
    Clearn locking which is older than time period.

    :param time_period in seconds
    """
    orm_transforms.clean_locking(time_period=time_period, session=session)


@transactional_session
def clean_next_poll_at(status, session=None):
    """
    Clearn next_poll_at.

    :param status: status of the transform
    """
    orm_transforms.clean_next_poll_at(status=status, session=session)


@read_session
def get_transform_input_output_maps(transform_id, input_coll_ids, output_coll_ids, log_coll_ids=[], session=None):
    """
    Get transform input output maps.

    :param transform_id: transform id.
    """
    contents = orm_contents.get_contents_by_request_transform(transform_id=transform_id, session=session)
    ret = {}
    for content in contents:
        map_id = content['map_id']
        if map_id not in ret:
            ret[map_id] = {'inputs_dependency': [], 'inputs': [], 'outputs': [], 'logs': [], 'others': []}

        """
        if content['coll_id'] in input_coll_ids:
            ret[map_id]['inputs'].append(content)
        elif content['coll_id'] in output_coll_ids:
            ret[map_id]['outputs'].append(content)
        elif content['coll_id'] in log_coll_ids:
            ret[map_id]['logs'].append(content)
        else:
            ret[map_id]['others'].append(content)
        """
        if content['content_relation_type'] == ContentRelationType.Input:
            ret[map_id]['inputs'].append(content)
        elif content['content_relation_type'] == ContentRelationType.InputDependency:
            ret[map_id]['inputs_dependency'].append(content)
        elif content['content_relation_type'] == ContentRelationType.Output:
            ret[map_id]['outputs'].append(content)
        elif content['content_relation_type'] == ContentRelationType.Log:
            ret[map_id]['logs'].append(content)
        else:
            ret[map_id]['others'].append(content)
    return ret


def release_inputs(to_release_inputs):
    update_contents = []
    for to_release in to_release_inputs:
        contents = orm_contents.get_input_contents(request_id=to_release['request_id'],
                                                   coll_id=to_release['coll_id'],
                                                   name=to_release['name'])
        for content in contents:
            if content['content_relation_type'] == ContentRelationType.InputDependency:
                update_content = {'content_id': content['content_id'],
                                  'substatus': to_release['substatus'],
                                  'status': to_release['status']}
                update_contents.append(update_content)
    return update_contents


def release_inputs_by_collection_old(to_release_inputs):
    update_contents = []
    for coll_id in to_release_inputs:
        to_release_contents = to_release_inputs[coll_id]
        if to_release_contents:
            to_release = to_release_contents[0]
            to_release_names_available = []
            to_release_names_fake_available = []
            to_release_names_final_failed = []
            to_release_names_missing = []
            for to_release_content in to_release_contents:
                if (to_release_content['status'] in [ContentStatus.Available]            # noqa: W503
                   or to_release_content['substatus'] in [ContentStatus.Available]):    # noqa: W503
                    to_release_names_available.append(to_release_content['name'])
                elif (to_release_content['status'] in [ContentStatus.FakeAvailable]            # noqa: W503
                     or to_release_content['substatus'] in [ContentStatus.FakeAvailable]):    # noqa: W503, E128
                    to_release_names_fake_available.append(to_release_content['name'])
                elif (to_release_content['status'] in [ContentStatus.FinalFailed]            # noqa: W503
                     or to_release_content['substatus'] in [ContentStatus.FinalFailed]):    # noqa: W503, E128
                    to_release_names_final_failed.append(to_release_content['name'])
                elif (to_release_content['status'] in [ContentStatus.Missing]            # noqa: W503
                      or to_release_content['substatus'] in [ContentStatus.Missing]):    # noqa: W503
                    to_release_names_missing.append(to_release_content['name'])
            contents = orm_contents.get_input_contents(request_id=to_release['request_id'],
                                                       coll_id=to_release['coll_id'],
                                                       name=None)

            for content in contents:
                if (content['content_relation_type'] == ContentRelationType.InputDependency):    # noqa: W503
                    if (content['status'] not in [ContentStatus.Available]                       # noqa: W503
                       and content['name'] in to_release_names_available):                          # noqa: W503
                        update_content = {'content_id': content['content_id'],
                                          'substatus': ContentStatus.Available,
                                          'status': ContentStatus.Available}
                        update_contents.append(update_content)
                    elif (content['status'] not in [ContentStatus.FakeAvailable]                     # noqa: W503
                          and content['name'] in to_release_names_fake_available):                        # noqa: W503
                        update_content = {'content_id': content['content_id'],
                                          'substatus': ContentStatus.FakeAvailable,
                                          'status': ContentStatus.FakeAvailable}
                        update_contents.append(update_content)
                    elif (content['status'] not in [ContentStatus.FinalFailed]                     # noqa: W503
                          and content['name'] in to_release_names_final_failed):                        # noqa: W503
                        update_content = {'content_id': content['content_id'],
                                          'substatus': ContentStatus.FinalFailed,
                                          'status': ContentStatus.FinalFailed}
                        update_contents.append(update_content)
                    elif (content['status'] not in [ContentStatus.Missing]                     # noqa: W503
                          and content['name'] in to_release_names_missing):                        # noqa: W503
                        update_content = {'content_id': content['content_id'],
                                          'substatus': ContentStatus.Missing,
                                          'status': ContentStatus.Missing}
                        update_contents.append(update_content)
    return update_contents


def release_inputs_by_collection(to_release_inputs, final=False):
    update_contents = []
    status_to_check = [ContentStatus.Available, ContentStatus.FakeAvailable, ContentStatus.FinalFailed, ContentStatus.Missing]
    for coll_id in to_release_inputs:
        to_release_contents = to_release_inputs[coll_id]
        if to_release_contents:
            to_release_status = {}
            for to_release_content in to_release_contents:
                if (to_release_content['status'] in status_to_check):
                    to_release_status[to_release_content['name']] = to_release_content['status']
                elif (to_release_content['substatus'] in status_to_check):
                    to_release_status[to_release_content['name']] = to_release_content['substatus']

            # print("to_release_status: %s" % str(to_release_status))

            contents = orm_contents.get_input_contents(request_id=to_release_contents[0]['request_id'],
                                                       coll_id=to_release_contents[0]['coll_id'],
                                                       name=None)
            # print("contents: %s" % str(contents))

            unfinished_contents_dict = {}
            for content in contents:
                if (content['content_relation_type'] == ContentRelationType.InputDependency):    # noqa: W503
                    if content['status'] not in status_to_check:
                        if content['name'] not in unfinished_contents_dict:
                            unfinished_contents_dict[content['name']] = []
                        content_short = {'content_id': content['content_id'], 'status': content['status']}
                        unfinished_contents_dict[content['name']].append(content_short)

            intersection_keys = to_release_status.keys() & unfinished_contents_dict.keys()
            intersection_keys = list(intersection_keys)
            logging.debug("release_inputs_by_collection(coll_id: %s): intersection_keys[:10]: %s" % (coll_id, str(intersection_keys[:10])))

            for name in intersection_keys:
                matched_content_status = to_release_status[name]
                matched_contents = unfinished_contents_dict[name]
                for matched_content in matched_contents:
                    if (matched_content['status'] != matched_content_status):
                        update_content = {'content_id': matched_content['content_id'],
                                          'substatus': matched_content_status,
                                          'status': matched_content_status}
                        update_contents.append(update_content)

    return update_contents


def poll_inputs_dependency_by_collection(unfinished_inputs):
    update_contents = []
    status_to_check = [ContentStatus.Available, ContentStatus.FakeAvailable, ContentStatus.FinalFailed, ContentStatus.Missing]
    for coll_id in unfinished_inputs:
        unfinished_contents = unfinished_inputs[coll_id]
        contents = orm_contents.get_input_contents(request_id=unfinished_contents[0]['request_id'],
                                                   coll_id=unfinished_contents[0]['coll_id'],
                                                   name=None)

        logging.debug("poll_inputs_dependency_by_collection(coll_id: %s): unfinished_contents[:10]: %s" % (coll_id, str(unfinished_contents[:10])))

        to_release_status = {}
        for content in contents:
            if (content['content_relation_type'] == ContentRelationType.Output):    # noqa: W503
                if content['status'] in status_to_check:
                    to_release_status[content['name']] = content['status']
                elif content['substatus'] in status_to_check:
                    to_release_status[content['name']] = content['substatus']

        unfinished_contents_dict = {}
        for content in unfinished_contents:
            if content['name'] not in unfinished_contents_dict:
                unfinished_contents_dict[content['name']] = []
            content_short = {'content_id': content['content_id'], 'status': content['status']}
            unfinished_contents_dict[content['name']].append(content_short)

        intersection_keys = to_release_status.keys() & unfinished_contents_dict.keys()
        intersection_keys = list(intersection_keys)
        logging.debug("poll_inputs_dependency_by_collection(coll_id: %s): intersection_keys[:10]: %s" % (coll_id, str(intersection_keys[:10])))

        for name in intersection_keys:
            matched_content_status = to_release_status[name]
            matched_contents = unfinished_contents_dict[name]
            for matched_content in matched_contents:
                if (matched_content['status'] != matched_content_status):
                    update_content = {'content_id': matched_content['content_id'],
                                      'substatus': matched_content_status,
                                      'status': matched_content_status}
                    update_contents.append(update_content)

        # if len(unfinished_contents_dict.keys()) < len(to_release_status.keys()):
        #     for name, content in unfinished_contents_dict.items():
        #         if name in to_release_status:
        #             matched_content_status = to_release_status[name]
        #             if (content['status'] != matched_content_status):
        #                 update_content = {'content_id': content['content_id'],
        #                                   'substatus': matched_content_status,
        #                                   'status': matched_content_status}
        #                 update_contents.append(update_content)
        # else:
        #     for name, status in to_release_status.items():
        #         if name in unfinished_contents_dict:
        #             matched_content = unfinished_contents_dict[name]
        #             if (matched_content['status'] != status):
        #                 update_content = {'content_id': matched_content['content_id'],
        #                                   'substatus': status,
        #                                   'status': status}
        #                 update_contents.append(update_content)

        # for content in unfinished_contents:
        #     if content['name'] in to_release_status:
        #         matched_content_status = to_release_status[content['name']]
        #         if (content['status'] != matched_content_status):
        #             update_content = {'content_id': content['content_id'],
        #                               'substatus': matched_content_status,
        #                               'status': matched_content_status}
        #             update_contents.append(update_content)

    return update_contents


def get_work_name_to_coll_map(request_id):
    tfs = orm_transforms.get_transforms(request_id=request_id)
    colls = orm_collections.get_collections(request_id=request_id)
    work_name_to_coll_map = {}
    for tf in tfs:
        if ('transform_metadata' in tf and tf['transform_metadata']
           and 'work_name' in tf['transform_metadata'] and tf['transform_metadata']['work_name']):  # noqa: W503
            work_name = tf['transform_metadata']['work_name']
            transform_id = tf['transform_id']
            if work_name not in work_name_to_coll_map:
                work_name_to_coll_map[work_name] = {'inputs': [], 'outputs': []}
            for coll in colls:
                if coll['transform_id'] == transform_id:
                    if coll['relation_type'] == CollectionRelationType.Input:
                        work_name_to_coll_map[work_name]['inputs'].append({'coll_id': coll['coll_id'], 'transform_id': coll['transform_id'],
                                                                           'workload_id': coll['workload_id'],
                                                                           'scope': coll['scope'], 'name': coll['name']})
                    elif coll['relation_type'] == CollectionRelationType.Output:
                        work_name_to_coll_map[work_name]['outputs'].append({'coll_id': coll['coll_id'], 'transform_id': coll['transform_id'],
                                                                            'workload_id': coll['workload_id'],
                                                                            'scope': coll['scope'], 'name': coll['name']})
    return work_name_to_coll_map



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2023


import logging
import json
import random
import socket
import threading
import time
import traceback
import stomp

from idds.common.plugin.plugin_base import PluginBase
from idds.common.utils import setup_logging, get_logger


setup_logging(__name__)
logging.getLogger("stomp").setLevel(logging.CRITICAL)


class MessagingListener(stomp.ConnectionListener):
    '''
    Messaging Listener
    '''
    def __init__(self, broker, output_queue, logger=None):
        '''
        __init__
        '''
        self.name = "MessagingListener"
        self.__broker = broker
        self.__output_queue = output_queue
        # self.logger = logging.getLogger(self.__class__.__name__)
        if logger:
            self.logger = logger
        else:
            self.logger = get_logger(self.__class__.__name__)

    def on_error(self, frame):
        '''
        Error handler
        '''
        self.logger.error('[broker] [%s]: %s', self.__broker, frame.body)

    def on_message(self, frame):
        self.logger.debug('[broker] [%s]: %s', self.__broker, frame.body)
        self.__output_queue.put(frame.body)
        pass


class MessagingSender(PluginBase, threading.Thread):
    def __init__(self, name="MessagingSender", logger=None, **kwargs):
        threading.Thread.__init__(self, name=name)
        super(MessagingSender, self).__init__(name=name, logger=logger, **kwargs)

        if logger:
            self.logger = logger
        self.graceful_stop = threading.Event()
        self.request_queue = None
        self.output_queue = None
        self.response_queue = None

        if not hasattr(self, 'channels'):
            raise Exception('"channels" is required but not defined.')
        self.channels = json.loads(self.channels)

        self.broker_timeout = 3600

        if not hasattr(self, 'timetolive'):
            self.timetolive = 12 * 3600 * 1000     # milliseconds
        else:
            self.timetolive = int(self.timetolive)

        self.conns = []

    def setup_logger(self, logger):
        if logger:
            self.logger = logger
        else:
            self.logger = get_logger(self.__class__.__name__)

    def set_logger(self, logger):
        self.logger = logger

    def get_logger(self):
        return self.logger

    def stop(self):
        self.graceful_stop.set()

    def set_request_queue(self, request_queue):
        self.request_queue = request_queue

    def set_output_queue(self, output_queue):
        self.output_queue = output_queue

    def set_response_queue(self, response_queue):
        self.response_queue = response_queue

    def connect_to_messaging_brokers(self, sender=True):
        channel_conns = {}
        for name in self.channels:
            channel = self.channels[name]
            if channel and 'brokers' in channel:
                brokers = channel['brokers']
                if type(brokers) in [list, tuple]:
                    pass
                else:
                    brokers = brokers.split(",")
                # destination = channel['destination']
                # username = channel['username']
                # password = channel['password']
                broker_timeout = channel['broker_timeout']

                broker_addresses = []
                for b in brokers:
                    try:
                        b, port = b.split(":")

                        addrinfos = socket.getaddrinfo(b, 0, socket.AF_INET, 0, socket.IPPROTO_TCP)
                        for addrinfo in addrinfos:
                            b_addr = addrinfo[4][0]
                            broker_addresses.append((b_addr, port))
                    except socket.gaierror as error:
                        self.logger.error('Cannot resolve hostname %s: %s' % (b, str(error)))

                self.logger.info("Resolved broker addresses for channel %s: %s" % (name, broker_addresses))

                timeout = broker_timeout

                conns = []
                for broker, port in broker_addresses:
                    conn = stomp.Connection12(host_and_ports=[(broker, port)],
                                              keepalive=True,
                                              heartbeats=(60000, 60000),     # one minute
                                              timeout=timeout)
                    conns.append(conn)
                channel_conns[name] = conns
            else:
                channel_conns[name] = None
        return channel_conns

    def disconnect(self, conns):
        for name in conns:
            if conns[name]:
                for conn in conns[name]:
                    try:
                        conn.disconnect()
                    except Exception:
                        pass

    def get_connection(self, destination):
        try:
            if destination not in self.conns:
                destination = 'default'
            if self.conns[destination]:
                conn = random.sample(self.conns[destination], 1)[0]
                queue_dest = self.channels[destination]['destination']
                username = self.channels[destination]['username']
                password = self.channels[destination]['password']
                if not conn.is_connected():
                    # conn.start()
                    conn.connect(username, password, wait=True)
                    return conn, queue_dest, destination
            else:
                return None, None, destination
        except Exception as error:
            self.logger.error("Failed to connect to message broker(will re-resolve brokers): %s" % str(error))

        self.disconnect(self.conns)

        try:
            self.conns = self.connect_to_messaging_brokers(sender=True)
            if destination not in self.conns:
                destination = 'default'
            conn = random.sample(self.conns[destination], 1)[0]
            queue_dest = self.channels[destination]['destination']
            if not conn.is_connected():
                conn.connect(self.username, self.password, wait=True)
                return conn, queue_dest, destination
        except Exception as error:
            self.logger.error("Failed to connect to message broker(will re-resolve brokers): %s" % str(error))

    def send_message(self, msg):
        destination = msg['destination'] if 'destination' in msg else 'default'
        conn, queue_dest, destination = self.get_connection(destination)

        if conn:
            self.logger.info("Sending message to message broker(%s): %s" % (destination, msg['msg_id']))
            self.logger.debug("Sending message to message broker(%s): %s" % (destination, json.dumps(msg['msg_content'])))
            conn.send(body=json.dumps(msg['msg_content']),
                      destination=queue_dest,
                      id='atlas-idds-messaging',
                      ack='auto',
                      headers={'persistent': 'true',
                               'ttl': self.timetolive,
                               'vo': 'atlas',
                               'msg_type': str(msg['msg_type']).lower()})
        else:
            self.logger.info("No brokers defined, discard(%s): %s" % (destination, msg['msg_id']))

    def execute_send(self):
        try:
            self.conns = self.connect_to_messaging_brokers(sender=True)
        except Exception as error:
            self.logger.error("Messaging sender throws an exception: %s, %s" % (error, traceback.format_exc()))

        while not self.graceful_stop.is_set():
            try:
                if not self.request_queue.empty():
                    msg = self.request_queue.get(False)
                    if msg:
                        self.send_message(msg)
                        self.response_queue.put(msg)
                else:
                    time.sleep(0.1)
            except Exception as error:
                self.logger.error("Messaging sender throws an exception: %s, %s" % (error, traceback.format_exc()))

    def run(self):
        try:
            self.execute_send()
        except Exception as error:
            self.logger.error("Messaging sender throws an exception: %s, %s" % (error, traceback.format_exc()))

    def __call__(self):
        self.run()


class MessagingReceiver(MessagingSender):
    def __init__(self, name="MessagingReceiver", logger=None, **kwargs):
        super(MessagingReceiver, self).__init__(name=name, logger=logger, **kwargs)
        self.listener = None
        self.receiver_conns = []

    def get_listener(self, broker):
        if self.listener is None:
            self.listener = MessagingListener(broker, self.output_queue, logger=self.logger)
        return self.listener

    def subscribe(self):
        self.receiver_conns = self.connect_to_messaging_brokers()

        for name in self.receiver_conns:
            for conn in self.receiver_conns[name]:
                self.logger.info('connecting to %s' % conn.transport._Transport__host_and_ports[0][0])
                conn.set_listener('message-receiver', self.get_listener(conn.transport._Transport__host_and_ports[0]))
                conn.connect(self.channels[name]['username'], self.channels[name]['password'], wait=True)
                conn.subscribe(destination=self.channels[name]['destination'], id='atlas-idds-messaging', ack='auto')

    def execute_subscribe(self):
        try:
            self.subscribe()
        except Exception as error:
            self.logger.error("Messaging receiver throws an exception: %s, %s" % (error, traceback.format_exc()))

        while not self.graceful_stop.is_set():
            has_failed_connection = False
            try:
                for name in self.receiver_conns:
                    for conn in self.receiver_conns[name]:
                        if not conn.is_connected():
                            conn.set_listener('message-receiver', self.get_listener(conn.transport._Transport__host_and_ports[0]))
                            # conn.start()
                            conn.connect(self.channels[name]['username'], self.channels[name]['password'], wait=True)
                            conn.subscribe(destination=self.channels[name]['destination'], id='atlas-idds-messaging', ack='auto')
                time.sleep(0.1)
            except Exception as error:
                self.logger.error("Messaging receiver throws an exception: %s, %s" % (error, traceback.format_exc()))
                has_failed_connection = True

            if has_failed_connection or len(self.receiver_conns) == 0:
                try:
                    # re-subscribe
                    self.disconnect(self.receiver_conns)
                    self.subscribe()
                except Exception as error:
                    self.logger.error("Messaging receiver throws an exception: %s, %s" % (error, traceback.format_exc()))

        self.logger.info('receiver graceful stop requested')

        self.disconnect(self.receiver_conns)

    def run(self):
        try:
            self.execute_subscribe()
        except Exception as error:
            self.logger.error("Messaging receiver throws an exception: %s, %s" % (error, traceback.format_exc()))

    def __call__(self):
        self.run()


class MessagingMessager(MessagingReceiver):
    def __init__(self, name="MessagingMessager", logger=None, **kwargs):
        super(MessagingMessager, self).__init__(name=name, logger=logger, **kwargs)

    def execute_send_subscribe(self):
        try:
            self.conns = self.connect_to_messaging_brokers(sender=True)
            self.subscribe()
        except Exception as error:
            self.logger.error("Messaging sender_subscriber throws an exception: %s, %s" % (error, traceback.format_exc()))

        while not self.graceful_stop.is_set():
            # send
            while True:
                try:
                    if not self.request_queue.empty():
                        msg = self.request_queue.get(False)
                        if msg:
                            self.send_message(msg)
                            if self.response_queue:
                                self.response_queue.put(msg)
                    else:
                        break
                except Exception as error:
                    self.logger.error("Messaging sender throws an exception: %s, %s" % (error, traceback.format_exc()))

            # subscribe
            has_failed_connection = False
            try:
                for name in self.receiver_conns:
                    for conn in self.receiver_conns[name]:
                        if not conn.is_connected():
                            conn.set_listener('message-receiver', self.get_listener(conn.transport._Transport__host_and_ports[0]))
                            # conn.start()
                            conn.connect(self.channels[name]['username'], self.channels[name]['password'], wait=True)
                            conn.subscribe(destination=self.channels[name]['destination'], id='atlas-idds-messaging', ack='auto')
            except Exception as error:
                self.logger.error("Messaging receiver throws an exception: %s, %s" % (error, traceback.format_exc()))
                has_failed_connection = True

            if has_failed_connection or len(self.receiver_conns) == 0:
                try:
                    # re-subscribe
                    self.disconnect(self.receiver_conns)
                    self.subscribe()
                except Exception as error:
                    self.logger.error("Messaging receiver throws an exception: %s, %s" % (error, traceback.format_exc()))

            time.sleep(0.1)

        self.logger.info('sender_receiver graceful stop requested')
        self.disconnect(self.conns)
        self.disconnect(self.receiver_conns)

    def run(self):
        try:
            self.execute_send_subscribe()
        except Exception as error:
            self.logger.error("Messaging receiver throws an exception: %s, %s" % (error, traceback.format_exc()))

    def __call__(self):
        self.run()



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2023


"""
operations related to Catalog(Collections and Contents).
"""


from idds.common import exceptions
from idds.common.constants import (CollectionType, CollectionStatus, CollectionLocking,
                                   CollectionRelationType, ContentStatus, ContentRelationType)
from idds.orm.base.session import read_session, transactional_session
from idds.orm import (transforms as orm_transforms,
                      collections as orm_collections,
                      contents as orm_contents,
                      messages as orm_messages)


@transactional_session
def get_collections_by_status(status, relation_type=CollectionRelationType.Input, time_period=None,
                              locking=False, bulk_size=None, to_json=False, session=None):
    """
    Get collections by status, relation_type and time_period or raise a NoObject exception.

    :param status: The collection status.
    :param relation_type: The relation_type of the collection to the transform.
    :param time_period: time period in seconds since last update.
    :param locking: Whether to retrieve unlocked files and lock them.
    :param to_json: return json format.

    :param session: The database session in use.

    :raises NoObject: If no collections are founded.

    :returns: list of Collections.
    """
    colls = orm_collections.get_collections_by_status(status=status, relation_type=relation_type, bulk_size=bulk_size,
                                                      time_period=time_period, locking=locking, to_json=to_json,
                                                      session=session)

    if locking:
        parameters = {'locking': CollectionLocking.Locking}
        for coll in colls:
            orm_collections.update_collection(coll_id=coll['coll_id'], parameters=parameters, session=session)
    return colls


@read_session
def get_collections(scope=None, name=None, request_id=None, workload_id=None, transform_id=None,
                    relation_type=None, to_json=False, session=None):
    """
    Get collections by scope, name, request_id and workload id.

    :param scope: scope of the collection.
    :param name: name the the collection.
    :param request_id: the request id.
    :param workload_id: The workload_id of the request.
    :param transform_id: The transform id related to this collection.
    :param relation_type: The relation between this collection and its transform,
                          such as Input, Output, Log and so on.
    :param to_json: return json format.
    :param session: The database session in use.

    :returns: dict of collections
    """
    collections = orm_collections.get_collections(scope=scope, name=name, request_id=request_id,
                                                  workload_id=workload_id, transform_id=transform_id,
                                                  to_json=to_json,
                                                  relation_type=relation_type, session=session)
    return collections


@read_session
def get_collections_by_request_ids(request_ids, session=None):
    """"
    Get collections by a list of request ids.

    :param request_ids: list of request ids.

    :return collections: list of collections.
    """
    return orm_collections.get_collections_by_request_ids(request_ids)


@transactional_session
def add_collection(request_id, workload_id, scope, name, coll_type=CollectionType.Dataset, transform_id=None,
                   relation_type=CollectionRelationType.Input, bytes=0, status=CollectionStatus.New,
                   total_files=0, new_files=0, processing_files=0, processed_files=0, retries=0,
                   expired_at=None, coll_metadata=None, session=None):
    """
    Add a collection.

    :param scope: The scope of the request data.
    :param name: The name of the request data.
    :param coll_type: The type of dataset as dataset or container.
    :param request_id: The request id.
    :param workload_id: The workload id.
    :param transform_id: The transform id related to this collection.
    :param relation_type: The relation between this collection and its transform,
                          such as Input, Output, Log and so on.
    :param bytes: The size of the collection.
    :param status: The status.
    :param total_files: Number of total files.
    :param retries: Number of retries.
    :param expired_at: The datetime when it expires.
    :param coll_metadata: The metadata as json.

    :raises DuplicatedObject: If a collection with the same name exists.
    :raises DatabaseException: If there is a database error.

    :returns: collection id.
    """
    orm_collections.add_collection(request_id=request_id, workload_id=workload_id,
                                   scope=scope, name=name, coll_type=coll_type,
                                   transform_id=transform_id, relation_type=relation_type,
                                   bytes=bytes, status=status, total_files=total_files,
                                   new_files=new_files, processing_files=processing_files,
                                   processed_files=processed_files, retries=retries, expired_at=expired_at,
                                   coll_metadata=coll_metadata, session=session)


@transactional_session
def update_collection(coll_id, parameters, msg=None, session=None):
    """
    update a collection.

    :param coll_id: the collection id.
    :param parameters: A dictionary of parameters.
    :param msg: messages.
    :param session: The database session in use.

    :raises NoObject: If no request is founded.
    :raises DatabaseException: If there is a database error.

    """
    orm_collections.update_collection(coll_id=coll_id, parameters=parameters, session=session)

    if msg:
        orm_messages.add_message(msg_type=msg['msg_type'],
                                 status=msg['status'],
                                 source=msg['source'],
                                 transform_id=msg['transform_id'],
                                 num_contents=msg['num_contents'],
                                 msg_content=msg['msg_content'],
                                 session=session)


@read_session
def get_collection(coll_id=None, transform_id=None, relation_type=None, to_json=False, session=None):
    """
    Get a collection or raise a NoObject exception.

    :param coll_id: The id of the collection.
    :param transform_id: The transform id related to this collection.
    :param relation_type: The relation between this collection and its transform,
                          such as Input, Output, Log and so on.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no request is founded.

    :returns: Collection.
    """
    return orm_collections.get_collection(coll_id=coll_id, transform_id=transform_id,
                                          relation_type=relation_type, to_json=to_json,
                                          session=session)


@transactional_session
def add_contents(contents, bulk_size=1000, session=None):
    """
    Add contents.

    :param contents: dict of contents.
    :param bulk_size: bulk per insert to db.
    :param session: session.

    :raises DuplicatedObject: If a collection with the same name exists.
    :raises DatabaseException: If there is a database error.

    :returns: content id.
    """
    return orm_contents.add_contents(contents=contents, bulk_size=bulk_size,
                                     session=session)


@transactional_session
def update_input_collection_with_contents(coll, parameters, contents, bulk_size=1000, session=None):
    """
    update a collection.

    :param coll_id: the collection id.
    :param parameters: A dictionary of parameters.
    :param contents: dict of contents.
    :param bulk_size: bulk per insert to db.
    :param session: The database session in use.

    :raises NoObject: If no request is founded.
    :raises DatabaseException: If there is a database error.

    :returns new contents
    """
    new_files = 0
    processed_files = 0
    avail_contents = orm_contents.get_contents(coll_id=coll['coll_id'], session=session)
    avail_contents_dict = {}
    for content in avail_contents:
        key = '%s:%s:%s:%s' % (content['scope'], content['name'], content['min_id'], content['max_id'])
        avail_contents_dict[key] = content
        if content['status'] in [ContentStatus.Mapped, ContentStatus.Mapped.value]:
            processed_files += 1
        if content['status'] in [ContentStatus.New, ContentStatus.New.value]:
            new_files += 1

    to_addes = []
    # to_updates = []
    for content in contents:
        key = '%s:%s:%s:%s' % (content['scope'], content['name'], content['min_id'], content['max_id'])
        if key in avail_contents_dict:
            """
            to_update = {'content_id': content['content_id'],
                         'status': content['status']}
            if 'bytes' in content:
                to_update['bytes'] = content['bytes']
            if 'md5' in content:
                to_update['md5'] = content['md5']
            if 'adler32' in content:
                to_update['adler32'] = content['adler32']
            if 'expired_at' in content:
                to_update['expired_at'] = content['expired_at']
            to_updates.append(to_updated)
            # not to do anything, no need to update
            """
            pass
        else:
            to_addes.append(content)

    # there are new files
    if to_addes:
        add_contents(to_addes, bulk_size=bulk_size, session=session)

    if 'total_files' in parameters:
        total_files = parameters['total_files']
    else:
        total_files = coll['total_files']
    parameters['processed_files'] = processed_files
    parameters['new_files'] = new_files
    if processed_files == total_files:
        parameters['status'] = CollectionStatus.Closed

    update_collection(coll['coll_id'], parameters, session=session)
    return to_addes


@transactional_session
def update_contents(parameters, session=None):
    """
    updatecontents.

    :param parameters: list of dictionary of parameters.
    :param session: The database session in use.

    :raises NoObject: If no content is founded.
    :raises DatabaseException: If there is a database error.

    """
    return orm_contents.update_contents(parameters, session=session)


@transactional_session
def update_content(content_id, parameters, session=None):
    """
    update a content.

    :param content_id: the content id.
    :param parameters: A dictionary of parameters.
    :param session: The database session in use.

    :raises NoObject: If no content is founded.
    :raises DatabaseException: If there is a database error.

    """
    return orm_contents.update_content(content_id, parameters, session=session)


@read_session
def get_contents(coll_scope=None, coll_name=None, request_id=None, workload_id=None, transform_id=None,
                 relation_type=None, status=None, to_json=False, session=None):
    """
    Get contents with collection scope, collection name, request id, workload id and relation type.

    :param coll_scope: scope of the collection.
    :param coll_name: name the the collection.
    :param request_id: the request id.
    :param workload_id: The workload_id of the request.
    :param transform_id: The transform id related to this collection.
    :param relation_type: The relation type between the collection and transform: input, outpu, logs and etc.
    :param to_json: return json format.
    :param session: The database session in use.

    :returns: list of contents
    """
    collections = get_collections(scope=coll_scope, name=coll_name, request_id=request_id,
                                  workload_id=workload_id, transform_id=transform_id,
                                  relation_type=relation_type, to_json=to_json, session=session)

    coll_ids = [coll['coll_id'] for coll in collections]
    if coll_ids:
        if relation_type is None:
            content_relation_type = None
        else:
            if relation_type == CollectionRelationType.Output:
                content_relation_type = ContentRelationType.Output
            elif relation_type == CollectionRelationType.Input:
                content_relation_type = ContentRelationType.Input
            elif relation_type == CollectionRelationType.Log:
                content_relation_type = ContentRelationType.Log
        rets = orm_contents.get_contents(coll_id=coll_ids, status=status, to_json=to_json,
                                         relation_type=content_relation_type, session=session)
    else:
        rets = []
    return rets


@read_session
def get_contents_by_request_transform(request_id=None, workload_id=None, transform_id=None, status=None, status_updated=False, session=None):
    """
    Get contents with request id, workload id and transform id.

    :param request_id: the request id.
    :param workload_id: The workload_id of the request.
    :param transform_id: The transform id related to this collection.
    :param session: The database session in use.

    :returns: list of contents
    """
    ret = orm_contents.get_contents_by_request_transform(request_id=request_id, transform_id=transform_id,
                                                         workload_id=workload_id, status=status,
                                                         status_updated=status_updated, session=session)
    return ret


@read_session
def get_contents_by_content_ids(content_ids, request_id=None, session=None):
    """
    Get content or raise a NoObject exception.

    :param request_id: request id.
    :param content_ids: list of content id.
    :param workload_id: workload id.

    :param session: The database session in use.

    :raises NoObject: If no content is founded.

    :returns: list of contents.
    """
    ret = orm_contents.get_contents_by_content_ids(content_ids=content_ids, request_id=request_id, session=session)
    return ret


@read_session
def get_contents_by_coll_id_status(coll_id, status=None, to_json=False, session=None):
    """
    Get contents or raise a NoObject exception.

    :param coll_id: Collection id.
    :param status: Content status or list of content status.
    :param to_json: return json format.
    :param session: The database session in use.

    :raises NoObject: If no content is founded.

    :returns: list of contents.
    """
    return orm_contents.get_contents(coll_id=coll_id, status=status, to_json=to_json, session=session)


@transactional_session
def register_output_contents(coll_scope, coll_name, contents, request_id=None, workload_id=None,
                             relation_type=CollectionRelationType.Output, session=None):
    """
    register contents with collection scope, collection name, request id, workload id and contents.

    :param coll_scope: scope of the collection.
    :param coll_name: name the the collection.
    :param request_id: the request id.
    :param workload_id: The workload_id of the request.
    :param contents: list of contents [{'scope': <scope>, 'name': <name>, 'min_id': min_id, 'max_id': max_id,
                                        'status': <status>, 'path': <path>}].
    :param session: The database session in use.
    """
    transform_ids = orm_transforms.get_transform_ids(request_id=request_id,
                                                     workload_id=workload_id,
                                                     session=session)

    if transform_ids:
        collections = orm_collections.get_collections(scope=coll_scope, name=coll_name, transform_id=transform_ids,
                                                      relation_type=relation_type, session=session)
    else:
        collections = []

    coll_def = "request_id=%s, workload_id=%s, coll_scope=%s" % (request_id, workload_id, coll_scope)
    coll_def += ", coll_name=%s, relation_type: %s" % (coll_name, relation_type)

    if len(collections) != 1:
        msg = "There should be only one collection matched. However there are %s collections" % len(collections)
        msg += coll_def
        raise exceptions.WrongParameterException(msg)

    coll_id = collections[0]['coll_id']

    keys = ['scope', 'name', 'min_id', 'max_id']
    for content in contents:
        ex_content = orm_contents.get_content(coll_id=coll_id, scope=content['scope'],
                                              name=content['name'], min_id=content['min_id'],
                                              max_id=content['max_id'], session=session)

        content_def = "scope: %s, name: %s, min_id: %s, max_id: %s" % (content['scope'],
                                                                       content['name'],
                                                                       content['min_id'],
                                                                       content['max_id'])

        if not ex_content:
            msg = "No matched content in collection(%s) with content(%s)" % (coll_def, content_def)
            raise exceptions.WrongParameterException(msg)

        for key in keys:
            if key in content:
                del content[key]
        content['content_id'] = ex_content['content_id']

    orm_contents.update_contents(contents, session=session)


@read_session
def get_match_contents(coll_scope, coll_name, scope, name, min_id=None, max_id=None,
                       request_id=None, workload_id=None, relation_type=None,
                       only_return_best_match=False, to_json=False, session=None):
    """
    Get matched contents with collection scope, collection name, scope, name, min_id, max_id,
    request id, workload id and only_return_best_match.

    :param coll_scope: scope of the collection.
    :param coll_name: name the the collection.
    :param scope: scope of the content.
    :param name: name of the content.
    :param min_id: min_id of the content.
    :param max_id: max_id of the content.
    :param request_id: the request id.
    :param workload_id: The workload_id of the request.
    :param only_return_best_match: only return best matched content if it's true.
    :param session: The database session in use.

    :returns: list of contents
    """
    transform_ids = orm_transforms.get_transform_ids(request_id=request_id,
                                                     workload_id=workload_id,
                                                     session=session)

    if transform_ids:
        collections = orm_collections.get_collections(scope=coll_scope, name=coll_name, transform_id=transform_ids,
                                                      relation_type=relation_type, session=session)
    else:
        collections = []

    coll_def = "request_id=%s, workload_id=%s, coll_scope=%s" % (request_id, workload_id, coll_scope)
    coll_def += ", coll_name=%s, relation_type: %s" % (coll_name, relation_type)

    if len(collections) != 1:
        msg = "There should be only one collection matched. However there are %s collections" % len(collections)
        msg += coll_def
        raise exceptions.WrongParameterException(msg)

    coll_id = collections[0]['coll_id']

    contents = orm_contents.get_match_contents(coll_id=coll_id, scope=scope, name=name, min_id=min_id, max_id=max_id,
                                               to_json=to_json, session=session)

    if not only_return_best_match:
        return contents

    if len(contents) == 1:
        return contents

    content = None
    for row in contents:
        if (not content) or (content['max_id'] - content['min_id'] > row['max_id'] - row['min_id']):
            content = row
    return [content]


@read_session
def get_content_status_statistics(coll_id=None, session=None):
    """
    Get statistics group by status

    :param coll_id: Collection id.
    :param session: The database session in use.

    :returns: statistics group by status, as a dict.
    """
    return orm_contents.get_content_status_statistics(coll_id=coll_id, session=session)


@transactional_session
def clean_locking(time_period=3600, session=None):
    """
    Clearn locking which is older than time period.

    :param time_period in seconds
    """
    orm_collections.clean_locking(time_period=time_period, session=session)


@transactional_session
def clean_next_poll_at(status, session=None):
    """
    Clearn next_poll_at.

    :param status: status of the collection
    """
    orm_collections.clean_next_poll_at(status=status, session=session)


@read_session
def get_output_content_by_request_id_content_name(request_id, content_scope, content_name, transform_id=None,
                                                  content_type=None, min_id=None, max_id=None, to_json=False,
                                                  session=None):
    """
    Get output content by request_id and content name

    :param request_id: requestn id.
    :param content_name: The name of the content.
    :param to_json: return json format.
    :param session: The database session in use.

    :returns: content of the output collection.
    """
    transform_ids = orm_transforms.get_transform_ids(request_id, session=session)

    found_transform_id = None
    if transform_ids:
        if len(transform_ids) == 1:
            found_transform_id = transform_ids[0]
        elif len(transform_ids) > 1 and transform_id is None:
            raise "Number of the transforms(%s) is bigger than 1 and transform id is not provided" % len(transform_ids)
        else:
            for tf_id in transform_ids:
                if tf_id == transform_id:
                    found_transform_id = tf_id
                    break

    coll_id = None
    if found_transform_id:
        coll_id = orm_collections.get_collection_id(transform_id=found_transform_id,
                                                    relation_type=CollectionRelationType.Output,
                                                    session=session)
    content = None
    if coll_id:
        content = orm_contents.get_content(coll_id=coll_id, scope=content_scope, name=content_name, content_type=content_type,
                                           min_id=min_id, max_id=max_id, to_json=to_json, session=session)
    return content


@read_session
def get_output_contents_by_request_id_status(request_id, name, content_status, limit, transform_id=None, to_json=False, session=None):
    """
    Get output content by request_id and content name

    :param request_id: requestn id.
    :param name: the content name.
    :param content_status: The content status.
    :param limit: limit number of contents.
    :param to_json: return json format.
    :param session: The database session in use.

    :returns: content of the output collection.
    """
    transform_ids = orm_transforms.get_transform_ids(request_id, session=session)

    found_transform_id = None
    if transform_ids:
        if len(transform_ids) == 1:
            found_transform_id = transform_ids[0]
        elif len(transform_ids) > 1 and transform_id is None:
            raise "Number of the transforms(%s) is bigger than 1 and transform id is not provided" % len(transform_ids)
        else:
            for tf_id in transform_ids:
                if tf_id == transform_id:
                    found_transform_id = tf_id
                    break

    coll_id = None
    if found_transform_id:
        coll_id = orm_collections.get_collection_id(transform_id=found_transform_id,
                                                    relation_type=CollectionRelationType.Output,
                                                    session=session)

    contents = []
    if coll_id:
        contents = orm_contents.get_contents(coll_id=coll_id, status=content_status, to_json=to_json, session=session)

    if name:
        new_contents = []
        for content in contents:
            if str(content['name']) == str(name):
                new_contents.append(content)
        contents = new_contents

    if contents and limit and len(contents) > limit:
        contents = contents[:limit]
    return contents


@read_session
def get_updated_transforms_by_content_status(request_id=None, transform_id=None, session=None):
    """
    Get updated transform ids by content status

    :param request_id: The Request id.
    :param transfomr_id: The transform id.

    :returns list
    """
    return orm_contents.get_updated_transforms_by_content_status(request_id=request_id, transform_id=transform_id, session=session)


@transactional_session
def update_contents_to_others_by_dep_id(request_id=None, transform_id=None, session=None):
    """
    Update contents to others by content_dep_id.

    :param request_id: The Request id.
    :param transfomr_id: The transform id.
    """
    return orm_contents.update_contents_to_others_by_dep_id(request_id=request_id, transform_id=transform_id, session=session)


@transactional_session
def update_contents_from_others_by_dep_id(request_id=None, transform_id=None, session=None):
    """
    Update contents from others by content_dep_id

    :param request_id: The Request id.
    :param transfomr_id: The transform id.
    """
    return orm_contents.update_contents_from_others_by_dep_id(request_id=request_id, transform_id=transform_id, session=session)


@transactional_session
def add_contents_update(contents, bulk_size=10000, session=None):
    """
    Add contents update.

    :param contents: dict of contents.
    :param session: session.

    :raises DuplicatedObject: If a collection with the same name exists.
    :raises DatabaseException: If there is a database error.

    :returns: content ids.
    """
    return orm_contents.add_contents_update(contents, bulk_size=bulk_size, session=session)


@transactional_session
def delete_contents_update(request_id=None, transform_id=None, session=None):
    """
    delete a content.

    :param session: The database session in use.

    :raises NoObject: If no content is founded.
    :raises DatabaseException: If there is a database error.
    """
    return orm_contents.delete_contents_update(request_id=request_id, transform_id=transform_id, session=session)


def get_contents_ext_maps():
    return orm_contents.get_contents_ext_maps()


@transactional_session
def add_contents_ext(contents, bulk_size=10000, session=None):
    """
    Add contents ext.

    :param contents: dict of contents.
    :param session: session.

    :raises DuplicatedObject: If a collection with the same name exists.
    :raises DatabaseException: If there is a database error.

    :returns: content ids.
    """
    return orm_contents.add_contents_ext(contents, bulk_size=bulk_size, session=session)


@transactional_session
def update_contents_ext(parameters, session=None):
    """
    update contents ext.

    :param parameters: list of dictionary of parameters.
    :param session: The database session in use.

    :raises NoObject: If no content is founded.
    :raises DatabaseException: If there is a database error.

    """
    return orm_contents.update_contents_ext(parameters, session=session)


@read_session
def get_contents_ext(request_id=None, transform_id=None, workload_id=None, coll_id=None, status=None, session=None):
    """
    Get content or raise a NoObject exception.

    :param request_id: request id.
    :param transform_id: transform id.
    :param workload_id: workload id.

    :param session: The database session in use.

    :raises NoObject: If no content is founded.

    :returns: list of contents.
    """
    return orm_contents.get_contents_ext(request_id=request_id, transform_id=transform_id, workload_id=workload_id,
                                         coll_id=coll_id, status=status, session=session)


@read_session
def get_contents_ext_ids(request_id=None, transform_id=None, workload_id=None, coll_id=None, status=None, session=None):
    """
    Get content or raise a NoObject exception.

    :param request_id: request id.
    :param transform_id: transform id.
    :param workload_id: workload id.

    :param session: The database session in use.

    :raises NoObject: If no content is founded.

    :returns: list of content ids.
    """
    return orm_contents.get_contents_ext_ids(request_id=request_id, transform_id=transform_id, workload_id=workload_id,
                                             coll_id=coll_id, status=status, session=session)


def combine_contents_ext(contents, contents_ext, with_status_name=False):
    return orm_contents.combine_contents_ext(contents, contents_ext, with_status_name=with_status_name)
